import collections
import errno
import itertools
import logging
import shutil

from pypeliner.sqlitedb import SqliteDb
import os

import pypeliner.helpers
import pypeliner.identifiers
import pypeliner.resources
import pypeliner.workflow


class NodeManager(object):
    """ Manages nodes in the underlying pipeline graph """

    def __init__(self, db, nodes_dir, temps_dir):
        self.db = db
        self.nodes_dir = nodes_dir
        self.temps_dir = temps_dir
        self.cached_chunks = dict()

    def retrieve_nodes(self, axes, base_node=None):
        if base_node is None:
            base_node = pypeliner.identifiers.Node()
        assert isinstance(base_node, pypeliner.identifiers.Node)
        if len(axes) == 0:
            yield base_node
        else:
            for chunk in self.retrieve_axis_chunks(axes[0], base_node):
                for node in self.retrieve_nodes(
                        axes[1:], base_node + pypeliner.identifiers.AxisInstance(axes[0], chunk)
                ):
                    yield node

    def get_chunks_filename(self, axis, node):
        return os.path.join(self.nodes_dir, node.subdir, axis + '_chunks')

    def retrieve_chunks(self, axes, node):
        assert isinstance(axes, tuple)
        if len(axes) == 0:
            yield ()
        else:
            for chunk in self.retrieve_axis_chunks(axes[0], node):
                for chunks_rest in self.retrieve_chunks(
                        axes[1:], node + pypeliner.identifiers.AxisInstance(axes[0], chunk)
                ):
                    yield (chunk,) + chunks_rest

    def retrieve_axis_chunks(self, axis, node):
        if (axis, node) not in self.cached_chunks:
            filename = self.db.get_temp_filename(axis, node)
            resource = pypeliner.resources.TempObjManager(self.db.file_storage, axis, node, filename)
            chunks = resource.get_obj()
            if chunks is None:
                chunks = (None,)
            self.cached_chunks[(axis, node)] = chunks
        return self.cached_chunks[(axis, node)]

    def store_chunks(self, axes, node, chunks, subset=None):
        if subset is None:
            subset = set([])
        if len(chunks) == 0:
            raise ValueError('must be at least one chunk per axis')

        def _convert_tuple(a):
            if isinstance(a, tuple):
                return a
            else:
                return tuple([a])

        chunks = [_convert_tuple(a) for a in chunks]
        if len(axes) != len(chunks[0]):
            raise ValueError('for multiple axis, chunks must be a tuple of the same length')
        for level in range(len(axes)):
            if level not in subset:
                continue
            for pre_chunks, level_chunks in itertools.groupby(sorted(chunks), lambda a: a[:level]):
                level_node = node
                for idx in range(level):
                    level_node += pypeliner.identifiers.AxisInstance(axes[idx], pre_chunks[idx])
                level_chunks = set([a[level] for a in level_chunks])
                self.store_axis_chunks(axes[level], level_node, level_chunks)

    def store_axis_chunks(self, axis, node, chunks):
        for chunk in chunks:
            new_node = node + pypeliner.identifiers.AxisInstance(axis, chunk)
        chunks = sorted(chunks)
        self.cached_chunks[(axis, node)] = chunks
        filename = self.db.get_temp_filename(axis, node)
        resource = pypeliner.resources.TempObjManager(self.db.file_storage, axis, node, filename)
        resource.finalize(chunks)

    def get_merge_inputs(self, axes, node, subset=None):
        if subset is None:
            subset = set([])
        subset = set(range(len(axes))).difference(subset)
        resources = self.get_chunks_resource(axes, node, subset)
        inputs = [resource.input for resource in resources]
        return inputs

    def get_split_outputs(self, axes, node, subset=None):
        if subset is None:
            subset = set([])
        resources = self.get_chunks_resource(axes, node, subset)
        outputs = [resource.output for resource in resources]
        return outputs

    def get_chunks_resource(self, axes, node, subset):
        if 0 in subset:
            filename = self.db.get_temp_filename(axes[0], node)
            yield pypeliner.resources.TempObjManager(self.db.file_storage, axes[0], node, filename)
        for level in range(1, len(axes)):
            if level not in subset:
                continue
            for level_node in self.retrieve_nodes(axes[:level], base_node=node):
                filename = self.db.get_temp_filename(axes[level], level_node)
                yield pypeliner.resources.TempObjManager(self.db.file_storage, axes[level], level_node, filename)

    def get_node_inputs(self, node):
        if len(node) >= 1:
            yield pypeliner.resources.Dependency(node[-1][0], node[:-1])


def _check_template(template, node):
    for axis, _ in node:
        if axis not in template:
            raise ValueError('filename template {} does not contain {{{}}}'.format(template, axis))


def resolve_user_filename(name, node, path_info):
    """ Resolve a filename based on user provided information """
    if path_info.fnames is not None:
        fname_key = tuple([a[1] for a in node])
        if len(fname_key) == 1:
            fname_key = fname_key[0]
        filename = path_info.fnames.get(fname_key)
    elif path_info.template is not None:
        _check_template(path_info.template, node)
        filename = path_info.template.format(**dict(node))
    else:
        _check_template(name, node)
        try:
            filename = name.format(**dict(node))
        except KeyError as e:
            raise ValueError('template {} contains {}'.format(name, e.args[0]))
    return filename


class UserFilenameCreator(object):
    """ Function object for creating user filenames from name node pairs """

    def __init__(self, path_info):
        self.path_info = path_info

    def __call__(self, name, node):
        return resolve_user_filename(name, node, self.path_info)

    def __repr__(self):
        return '{0}.{1}({2})'.format(
            UserFilenameCreator.__module__,
            UserFilenameCreator.__name__,
            repr(self.path_info))


class TempFilenameCreator(object):
    """ Function object for creating filenames from name node pairs """

    def __init__(self, file_dir=''):
        self.file_dir = file_dir

    def __call__(self, name, node):
        return os.path.join(self.file_dir, node.subdir, name)

    def __repr__(self):
        return '{0}.{1}({2})'.format(
            TempFilenameCreator.__module__,
            TempFilenameCreator.__name__,
            self.file_dir)


class WorkflowDatabase(object):
    def __init__(self, temps_dir, workflow_dir, logs_dir, file_storage, job_shelf, path_info, instance_subdir):
        self.file_storage = file_storage
        self.job_shelf = job_shelf
        self.path_info = path_info
        self.instance_subdir = instance_subdir
        self.envs_dir = os.path.join(workflow_dir, 'envs')
        self.nodes_dir = os.path.join(workflow_dir, 'nodes', instance_subdir)
        self.temps_dir = os.path.join(temps_dir, instance_subdir)
        self.nodemgr = NodeManager(self, self.nodes_dir, self.temps_dir)
        self.logs_dir = os.path.join(logs_dir, instance_subdir)

    def get_user_filename_creator(self, name, axes, fnames=None, template=None):
        arg_path_info = pypeliner.workflow.UserPathInfo(fnames=fnames, template=template)
        if (name, axes) in self.path_info:
            user_path_info = self.path_info[(name, axes)]
            if arg_path_info.fnames is not None and arg_path_info.template is not None and arg_path_info != user_path_info:
                raise Exception('{} doesnt equal {}'.format(arg_path_info, user_path_info))
        else:
            user_path_info = arg_path_info
        return UserFilenameCreator(user_path_info)

    def get_user_filename(self, name, node, fnames=None, template=None):
        return self.get_user_filename_creator(name, node.axes, fnames=fnames, template=template)(name, node)

    def get_temp_filename_creator(self):
        return TempFilenameCreator(file_dir=self.temps_dir)

    def get_temp_filename(self, name, node):
        if os.path.isabs(name):
            raise Exception('name {} is an absolute path'.format(name))
        return self.get_temp_filename_creator()(name, node)


class PipelineLockedError(Exception):
    pass


class WorkflowDatabaseFactory(object):
    def __init__(self, temps_dir, workflow_dir, logs_dir, file_storage):
        self.temps_dir = temps_dir
        self.workflow_dir = workflow_dir
        self.logs_dir = logs_dir
        pypeliner.helpers.makedirs(self.workflow_dir)
        self.file_storage = file_storage
        self.job_shelf_filename = os.path.join(self.workflow_dir, 'jobs.db')
        self.lock_directories = list()

    def create(self, path_info, instance_subdir):
        self._add_lock(instance_subdir)
        db = WorkflowDatabase(
            self.temps_dir, self.workflow_dir, self.logs_dir, self.file_storage,
            self.job_shelf, path_info, instance_subdir)
        return db

    def _add_lock(self, instance_subdir):
        lock_directory = os.path.join(self.workflow_dir, 'locks', instance_subdir, '_lock')
        try:
            pypeliner.helpers.makedirs(os.path.join(lock_directory, os.path.pardir))
            os.mkdir(lock_directory)
        except OSError as e:
            if e.errno == errno.EEXIST:
                raise PipelineLockedError('Pipeline already running, remove {0} to override'.format(lock_directory))
            else:
                raise
        self.lock_directories.append(lock_directory)

    def __enter__(self):
        self.job_shelf = SqliteDb(self.job_shelf_filename)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.job_shelf.close()
        for lock_directory in self.lock_directories:
            try:
                os.rmdir(lock_directory)
            except:
                logging.exception('unable to unlock ' + lock_directory)



import logging
import subprocess

import pypeliner.helpers
import time


class QEnv(object):
    """ Paths to qsub, qstat, qdel """

    def __init__(self):
        self.qsub_bin = pypeliner.helpers.which('qsub')
        self.qstat_bin = pypeliner.helpers.which('qstat')
        self.qacct_bin = pypeliner.helpers.which('qacct')
        self.qdel_bin = pypeliner.helpers.which('qdel')


class LsfEnv(object):
    """ Paths to qsub, qstat, qdel in LSF"""

    def __init__(self):
        self.qsub_bin = pypeliner.helpers.which('bsub')
        self.qstat_bin = pypeliner.helpers.which('bjobs')
        self.qacct_bin = pypeliner.helpers.which('bhist')
        self.qdel_bin = pypeliner.helpers.which('bkill')

class SlurmEnv(object):
    """ Paths to qsub, qstat, qdel in LSF"""

    def __init__(self):
        self.qsub_bin = pypeliner.helpers.which('sbatch')
        self.qstat_bin = pypeliner.helpers.which('squeue')
        self.qacct_bin = pypeliner.helpers.which('sacct')
        self.qdel_bin = pypeliner.helpers.which('scancel')


class QstatError(Exception):
    pass


class QstatJobStatus(object):
    """ Class representing statuses retrieved using qstat """

    def __init__(self, qenv, qstat_period=30, max_qstat_failures=10):
        self.qenv = qenv
        self.qstat_period = qstat_period
        self.max_qstat_failures = max_qstat_failures
        self.qstat_attempt_time = None
        self.cached_job_status = None
        self.qstat_failures = 0
        self.qstat_time = None
        self.logger = logging.getLogger('pypeliner.execqueue')

    def update(self, jobs):
        """ Update cached job status after sleeping for remainder of polling time.
        """
        if self.qstat_attempt_time is not None:
            time_since_attempt = time.time() - self.qstat_attempt_time
            sleep_time = max(0, self.qstat_period - time_since_attempt)
            time.sleep(sleep_time)
        self.qstat_attempt_time = time.time()
        try:
            self.cached_job_status = self.get_qstat_job_status(jobs)
            self.qstat_failures = 0
            self.qstat_time = time.time()
        except:
            self.logger.exception('Unable to qstat')
            self.qstat_failures += 1
        if self.qstat_failures >= self.max_qstat_failures:
            raise QstatError('too many consecutive qstat failures')

    def finished(self, qsub_job_id, qsub_time):
        """ Query whether job is finished.

        Args:
            qsub_job_id (str): job id returned by qsub
            qsub_time (float): time job was submitted

        Returns:
            bool: job finished, None for no information
        """
        if self.cached_job_status is None:
            return False

        if qsub_time >= self.qstat_time:
            return False

        return qsub_job_id not in self.cached_job_status

    def errors(self, qsub_job_id):
        """ Query whether job is in error state.

        Returns:
            bool: job finished, None for no information
        """
        if self.cached_job_status is None:
            return None

        return 'e' in self.cached_job_status.get(qsub_job_id, '').lower()

    def get_qstat_job_status(self, jobs):
        """ Run qstat to obtain job statues.

        Returns:
            dict: job id to job status dictionary
        """
        job_status = dict()

        for line in subprocess.check_output([self.qenv.qstat_bin]).split('\n'):
            row = line.split()
            try:
                qsub_job_id = row[0]
                status = row[4].lower()
                job_status[qsub_job_id] = status
            except IndexError:
                continue
        return job_status


class QacctError(Exception):
    pass


class QacctWrapper(object):
    def __init__(self, qenv, job_id, qacct_stdout_filename, qacct_stderr_filename, max_qacct_failures=100):
        self.qenv = qenv
        self.job_id = job_id
        self.qacct_stdout_filename = qacct_stdout_filename
        self.qacct_stderr_filename = qacct_stderr_filename
        self.max_qacct_failures = max_qacct_failures
        self.qacct_failures = 0

        self.results = None

    def parse_qacct(self):
        """parse job completion status report"""
        qacct_results = dict()

        with open(self.qacct_stdout_filename, 'r') as qacct_file:
            for line in qacct_file:
                try:
                    key, value = line.split()
                except ValueError:
                    continue
                qacct_results[key] = value
        return qacct_results

    def check(self):
        """ Run qacct to obtain finished job info.
        """
        try:
            with open(self.qacct_stdout_filename, 'w') as qacct_stdout, \
                    open(self.qacct_stderr_filename, 'w') as qacct_stderr:
                subprocess.check_call([self.qenv.qacct_bin, '-j', self.job_id], stdout=qacct_stdout,
                                      stderr=qacct_stderr)
        except subprocess.CalledProcessError:
            self.qacct_failures += 1
            if self.qacct_failures > self.max_qacct_failures:
                raise QacctError()
            else:
                return None

        self.results = self.parse_qacct()


class LsfacctWrapper(QacctWrapper):
    """class to check job status after completion in LSF"""

    def parse_qacct(self):
        """parse job completion status report"""
        qacct_results = dict()

        with open(self.qacct_stdout_filename, 'r') as qacct_file:
            for line in qacct_file:
                line = line.strip()
                if 'Done successfully' in line and 'Post job process' not in line:
                    qacct_results['exit_code'] = 0
                if 'Exited with exit code' in line:
                    line = line[line.find('Exited with exit code'):].split('.')[0]
                    line = line.replace('Exited with exit code', '')
                    qacct_results['exit_code'] = line
                if 'MAX MEM' in line:
                    line = line.split(';')[0].split(':')
                    assert line[0] == 'MAX MEM'
                    qacct_results['maxvmem'] = line[1]
        return qacct_results


class SlurmacctWrapper(QacctWrapper):
    """class to check job status after completion in LSF"""

    def parse_qacct(self):
        """parse job completion status report"""
        qacct_results = dict()

        with open(self.qacct_stdout_filename, 'r') as qacct_file:
            lines = [line.strip().split('|') for line in qacct_file]
            assert len(lines) == 2
            header = {v: i for i, v in enumerate(lines[0])}
            exit_code = lines[1][header['ExitCode']]
            status = lines[1][header['State']]
            qacct_results[exit_code] = exit_code
            assert status == 'COMPLETED' or status == 'FAILED'
        return qacct_results


class QsubWrapper(object):
    """class to submit jobs in SGE"""

    def __init__(self, qenv, ctx, name, script_filename, native_spec, job_stdout_filename,
                 job_stderr_filename, submit_stdout_filename, submit_stderr_filename):
        self.qsub_job_id = None
        self.qenv = qenv
        self.ctx = ctx
        self.name = name
        self.script_filename = script_filename
        self.native_spec = native_spec
        self.job_stdout_filename = job_stdout_filename
        self.job_stderr_filename = job_stderr_filename
        self.submit_stdout_filename = submit_stdout_filename
        self.submit_stderr_filename = submit_stderr_filename
        self.submit_command = self.create_submit_command()

    def create_submit_command(self):
        """generate the job submission command"""
        qsub = [self.qenv.qsub_bin]
        qsub += self.native_spec.format(**self.ctx).split()
        qsub += ['-N', pypeliner.execqueue.utils.qsub_format_name(self.name)]
        qsub += ['-o', self.job_stdout_filename]
        qsub += ['-e', self.job_stderr_filename]
        qsub += [self.script_filename]
        return qsub

    def extract_job_id(self):
        """parse the job id from job submission command output"""
        with open(self.submit_stdout_filename, 'r') as submit_stdout:
            return submit_stdout.readline().rstrip().replace('Your job ', '').split(' ')[0]

    def submit_job(self):
        """submit job and return job id"""
        try:
            with open(self.submit_stdout_filename, 'w') as submit_stdout, \
                    open(self.submit_stderr_filename, 'w') as submit_stderr:
                subprocess.check_call(self.submit_command, stdout=submit_stdout, stderr=submit_stderr)
        except Exception as e:
            raise pypeliner.execqueue.base.SubmitError('submit error ' + str(e))

        return self.extract_job_id()


class LsfsubWrapper(QsubWrapper):
    """class to submit jobs in LSF"""

    def create_submit_command(self):
        """generate the job submission command"""
        qsub = [self.qenv.qsub_bin]
        qsub += self.native_spec.format(**self.ctx).split()
        qsub += ['-J', pypeliner.execqueue.utils.qsub_format_name(self.name)]
        qsub += ['-o', self.job_stdout_filename]
        qsub += ['-eo', self.job_stderr_filename]
        qsub += [self.script_filename]
        return qsub

    def extract_job_id(self):
        """parse the job id from job submission command output"""
        with open(self.submit_stdout_filename, 'r') as submit_stdout:
            return submit_stdout.readline().rstrip().replace("<", "\t").replace(">", '\t').split('\t')[1]


class SlurmsubWrapper(QsubWrapper):
    """class to submit jobs in LSF"""

    def create_submit_command(self):
        """generate the job submission command"""
        qsub = [self.qenv.qsub_bin]
        qsub += self.native_spec.format(**self.ctx).split()
        qsub += ['-J', pypeliner.execqueue.utils.qsub_format_name(self.name)]
        qsub += ['-o', self.job_stdout_filename]
        qsub += ['-e', self.job_stderr_filename]
        qsub += [self.script_filename]
        return qsub

    def extract_job_id(self):
        """parse the job id from job submission command output"""
        with open(self.submit_stdout_filename, 'r') as submit_stdout:
            jobid = submit_stdout.readline().rstrip().replace("Submitted batch job", "")
            return int(jobid)



import logging
import random
import string


class AwsLoggingFilter(logging.Filter):
    """
    filter out AWS logging output from pypeliner log
    """
    def __init__(self):
        self.filter_keywords = [
            'botocore', 'boto3', 'urllib3', 's3transfer'
        ]

    def filter(self, rec):
        """
        filter a log record or not
        :param rec: log record
        :type rec:
        :return: True if filtered else False
        :rtype: bool
        """
        logname = rec.name.split('.')[0]
        if logname in self.filter_keywords:
            return False
        return True


def set_aws_logging_filters():
    """
    add aws filter to root logging handlers
    """
    for handler in logging.root.handlers:
        handler.addFilter(AwsLoggingFilter())


def random_string(length):
    """
    generate a random string of length
    :param length: length of string
    :type length: int
    :return: randomly generated string
    :rtype: str
    """
    return ''.join(random.choice(string.ascii_lowercase) for _ in range(length))



import pypeliner.commandline
import pypeliner.jobs


class UserPathInfo(object):
    def __init__(self, template=None, fnames=None):
        self.template = template
        self.fnames = fnames

    def __eq__(self, other):
        return self.template == other.template and self.fnames == other.fnames

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return '{0}.{1}({2}, {3})'.format(
            UserPathInfo.__module__,
            UserPathInfo.__name__,
            repr(self.template),
            repr(self.fnames))


parent_ctx = None


class Workflow(object):
    """ Contaner for a set of jobs making up a single workflow.

    """

    def __init__(self, ctx=None, default_ctx=None, default_sandbox=None):
        self.ctx = {
            'mem': 4,
            'num_retry': 3,
            'mem_retry_factor': 2,
            'ncpus': 1,
            'disk': 8
        }
        if parent_ctx is not None:
            self.ctx.update(parent_ctx)
        if ctx is not None:
            self.ctx.update(ctx)
        ## TODO: remove default_ctx at some point?
        if default_ctx is not None:
            DeprecationWarning("default_ctx will be replaced by ctx starting with v0.6")
            self.ctx.update(default_ctx)
        self.default_sandbox = default_sandbox
        self.job_definitions = dict()
        self.path_info = dict()

    @property
    def empty(self):
        return len(self.job_definitions) == 0

    def set_filenames(self, name, *axes, **kwargs):
        """ Set the filename for a
        """
        user_file_id = (name, axes)
        if user_file_id in self.path_info:
            raise ValueError(
                'Filename for {} with axes {} already set to {}'.format(name, axes, self.path_info[user_file_id])
            )
        self.path_info[user_file_id] = UserPathInfo()
        if 'fnames' in kwargs:
            self.path_info[user_file_id].fnames = kwargs['fnames']
        elif 'template' in kwargs:
            self.path_info[user_file_id].template = kwargs['template']
        elif 'filename' in kwargs:
            self.path_info[user_file_id].template = kwargs['filename']
        else:
            raise ValueError('One of fnames, template, or filename must be set')

    def setobj(self, obj=None, value=None, axes=()):
        """ Set a managed temp object with a specified value.

        :param obj: managed object to be set with a given value
        :type obj: :py:class:`pypeliner.managed.TempOutputObj`
        :param value: value to set
        :param axes: axes on which to perform operation.  If the axes argument is identical
                     to the axes of the object, the setting of the object occurs once per
                     axis chunk.  If the axes argument has length one less than the axes of
                     of obj, this setobj is a split operation that defines the additional
                     axis in obj.
        :type axes: tuple

        This function is most useful for tracking changes to small objects and parameters.
        Set the object to a given value using this function.  Then use the managed version
        of the object in calls to transform, and pypeliner will only trigger a rerun if the
        value of the object has changed in a subsequent run.

        """
        job_ctx = self.ctx.copy()
        job_ctx['local'] = True
        name = '_'.join(('setobj', str(obj.name)) + obj.axes)
        if name in self.job_definitions:
            raise ValueError('Object {} axes {} already set'.format(obj.name, repr(obj.axes)))

        self.job_definitions[name] = pypeliner.jobs.SetObjDefinition(name, axes, job_ctx, obj, value)

    def commandline(self, name='', axes=(), ctx=None, args=None, kwargs=None, sandbox=None):
        """ Add a command line based transform to the pipeline

        This call is equivalent to::

            self.transform(name, axes, ctx, commandline.execute, None, *args)

        See :py:func:`pypeliner.scheduler.transform`

        """
        if ctx is None:
            ctx = {}

        self.transform(
            name=name, axes=axes, ctx=ctx, func=pypeliner.commandline.execute,
            args=args, kwargs=kwargs, sandbox=None
        )

    def transform(self, name='', axes=(), ctx=None, func=None, ret=None, args=None, kwargs=None, sandbox=None):
        """ Add a transform to the pipeline.  A transform defines a job that uses the
        provided python function ``func`` to take input dependencies and create/update
        output dependents.

        :param name: unique name of the job, used to identify the job in logs and when
                     submitting instances to the exec queue
        :param axes: axes of the job.  defines the axes on which the job will operate.  A
                     job with an empty list for the axes has a single instance.  A job with
                     one axis in the axes list will have as many instances as were defined
                     for that axis by the split that is responsible for that axis.
        :param ctx: context of the job as a dictionary of key, value pairs.  The context
                    is given to the exec queue and provides a way of communicating jobs
                    specific requirements such as memory and cpu usage.  Setting
                    ``ctx['local'] = True`` will result in the job being run locally on
                    the calling machine even when a cluster is being used.
        :param func: The function to call for this job.
        :param ret: The return value
        :param args: The list of positional arguments to be used for the function call.
        :param kwargs: The list of keyword arguments to be used for the function call.

        Any value in args or kwargs that is an instance of
        :py:class:`pypeliner.managed.Managed` will be resolved to a pipeline managed
        file or object at runtime.  See :py:mod:`pypeliner.managed`.

        Acceptable values given for ``ret`` are restricted to a subset of
        :py:class:`pypeliner.managed.Managed` derived classes that represent output
        objects.  The return value of ``func`` will be stored and used by the pipelining
        system according to the specific details of the :py:class:`pypeliner.managed.Managed`
        derived class.

        """
        if not name:
            raise ValueError("job name not specified")
        job_ctx = self.ctx.copy()
        if ctx is not None:
            job_ctx.update(ctx)
        if name in self.job_definitions:
            raise ValueError('Job already defined')

        if sandbox is None:
            sandbox = self.default_sandbox
        self.job_definitions[name] = pypeliner.jobs.JobDefinition(
            name, axes, job_ctx, func, pypeliner.jobs.CallSet(ret=ret, args=args, kwargs=kwargs), sandbox=sandbox
        )

    def subworkflow(self, name='', axes=(), ctx=None, func=None, args=None, kwargs=None):
        """ Add a sub workflow to the pipeline.  A sub workflow is a set of jobs that
        takes the input dependencies and creates/updates output dependents.  The python
        function ``func`` should return a workflow object containing the set of jobs.

        :param name: unique name of the job, used to identify the job in logs and when
                     submitting instances to the exec queue
        :param axes: axes of the job.  defines the axes on which the job will operate.  A
                     job with an empty list for the axes has a single instance.  A job with
                     one axis in the axes list will have as many instances as were defined
                     for that axis by the split that is responsible for that axis.
        :param ctx: context of the subworkflow job as a dictionary of key, value pairs.  The context
                    is given to the exec queue and provides a way of communicating jobs
                    specific requirements such as memory and cpu usage.  Setting
                    ``ctx['local'] = True`` will result in the job being run locally on
                    the calling machine even when a cluster is being used.
        :param func: The function to call for this job.
        :param args: The list of positional arguments to be used for the function call.
        :param kwargs: The list of keyword arguments to be used for the function call.

        Any value in args or kwargs that is an instance of
        :py:class:`pypeliner.managed.Managed` will be resolved to a pipeline managed
        file or object at runtime.  See :py:mod:`pypeliner.managed`.

        """
        if not name:
            raise ValueError("subworkflow name not specified")
        job_ctx = self.ctx.copy()
        if ctx is not None:
            job_ctx.update(ctx)
        if name in self.job_definitions:
            raise ValueError('Job already defined')
        ret = pypeliner.managed.OutputWorkflow(name + '_workflow_definition', *axes)
        self.job_definitions[name] = pypeliner.jobs.SubWorkflowDefinition(
            name, axes, job_ctx, func, pypeliner.jobs.CallSet(ret=ret, args=args, kwargs=kwargs)
        )

    def _create_job_instances(self, graph, db):
        """ Create job instances from job definitions given resource and node managers,
        and a log directory.
        """
        for job_def in self.job_definitions.values():
            if hasattr(job_def, 'sandbox') and job_def.sandbox is not None:
                job_def.sandbox.create_conda_env(db.envs_dir)

        for job_def in self.job_definitions.values():
            for job_inst in job_def.create_job_instances(graph, db):
                yield job_inst



import logging
import os
import shutil
import subprocess
import sys
import tempfile
import time
import traceback

import dill as pickle
import pypeliner.helpers
import pypeliner.containerize


class Delegator(object):
    def __init__(self, job, prefix, modules):
        self.job = job
        self.before_filename = prefix + ".before"
        self.after_filename = prefix + ".after"
        self.syspaths = [os.path.dirname(os.path.abspath(module.__file__)) for module in modules]

    def cleanup(self):
        pypeliner.helpers.saferemove(self.before_filename)
        pypeliner.helpers.saferemove(self.after_filename)

    def _waitfile(self, filename):
        waittime = 1
        while waittime < 100:
            if os.path.exists(filename):
                return
            if waittime >= 4:
                logging.getLogger('pypeliner.delegator').warn(
                    'waiting {0}s for {1} to appear'.format(waittime, filename))
            time.sleep(waittime)
            waittime *= 2

    def initialize(self):
        self.cleanup()
        with open(self.before_filename, 'wb') as before:
            pickle.dump(self.job, before)
        command = ['pypeliner_delegate', self.before_filename, self.after_filename] + self.syspaths
        # local tasks like setobj need to run without docker args,
        # since they might already be running inside docker
        if not self.job.ctx.get('local'):
            command = pypeliner.containerize.containerize_args(*command)
        return command

    def finalize(self):
        self._waitfile(self.after_filename)
        if not os.path.exists(self.after_filename):
            return None
        self.job = None
        with open(self.after_filename, 'rb') as after:
            self.job = pickle.load(after)
        self.cleanup()
        if self.job:
            for logrecord in self.job.log_records:
                logging.getLogger().handle(logrecord)
        return self.job


def call_external(obj):
    try:
        tmp_dir = tempfile.mkdtemp()
        dgt = Delegator(obj, tmp_dir, [sys.modules[obj.__module__]])
        subprocess.check_call(dgt.initialize())
        return dgt.finalize()
    finally:
        try:
            shutil.rmtree(tmp_dir)
        except OSError as exc:
            if exc.errno != 2:
                raise


def main():
    job_logger = None

    try:
        job = None
        before_filename = sys.argv[1]
        after_filename = sys.argv[2]

        delegator_dir = os.path.dirname(before_filename)

        if not os.path.exists(delegator_dir) and pypeliner.helpers.running_in_docker():
            if delegator_dir:
                raise Exception(
                    "Did you forget to add your working directory to docker mounts?"
                    " {} is not available inside docker, delegator will "
                    "fail.".format(delegator_dir)
                )

        def not_pypeliner_path(a):
            if os.path.exists(a) and os.path.samefile(a, os.path.dirname(__file__)):
                return False
            return True

        sys.path = list(filter(not_pypeliner_path, sys.path))
        sys.path.extend(sys.argv[3:])

        with open(before_filename, 'rb') as before:
            job = pickle.load(before)
        if job is None:
            raise ValueError('no job data in ' + before_filename)
        if not job.version == pypeliner.__version__:
            logging.getLogger('pypeliner.delegator').warn(
                'mismatching pypeliner versions, '
                'running {} on head node and {} on compute node'.format(job.version, pypeliner.__version__))

        job_logger = pypeliner.helpers.RemoteLogger()
        logging.getLogger().addHandler(job_logger.log_handler)
        job()
    except:
        sys.stderr.write(traceback.format_exc())
    finally:
        if job_logger:
            job.log_records = job_logger.log_records
        with open(after_filename, 'wb') as after:
            pickle.dump(job, after)


if __name__ == "__main__":
    main()



from __future__ import print_function

import logging
import os

import datetime
import dill as pickle
import pypeliner.execqueue.base
import pypeliner.helpers
import pypeliner.tests.jobs
import time
import yaml

from .batchclient import BatchClient


class AzureJobQueue(object):
    """ Azure batch job queue.
    """

    def __init__(self, config_filename=None, **kwargs):
        batch_account_url = os.environ['AZURE_BATCH_URL']
        client_id = os.environ['CLIENT_ID']
        tenant_id = os.environ['TENANT_ID']
        secret_key = os.environ['SECRET_KEY']

        with open(config_filename) as f:
            self.config = yaml.safe_load(f)

        storage_account_name = self.config['pypeliner_storage_account']
        storage_keys = pypeliner.helpers.GlobalState.get('azure_storage_keys', {})
        account_token = storage_keys.get(storage_account_name)

        self.batch_client = BatchClient(
            client_id, tenant_id, secret_key, batch_account_url,
            self.config, storage_account_token=account_token
        )

        self.logger = self.batch_client.logger
        self.run_id = self.batch_client.run_id

        self.no_delete_pool = self.config.get('no_delete_pool', False)
        self.no_delete_job = self.config.get('no_delete_job', False)

        self.job_names = {}
        self.job_task_ids = {}
        self.job_temps_dir = {}
        self.job_blobname_prefix = {}

        self.mapping_from_tasks_to_job = {}

        self.most_recent_transition_time = None
        self.completed_task_ids = set()
        self.running_task_ids = set()

    debug_filenames = {
        'job stderr': 'stderr.txt',
        'job stdout': 'stdout.txt',
        'upload err': 'fileuploaderr.txt',
        'upload out': 'fileuploadout.txt',
    }

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.logger.info('tear down')
        if not self.no_delete_job:
            self.batch_client.delete_jobs(ignore_exception=True)
        if not self.no_delete_pool:
            self.batch_client.delete_pools(ignore_exception=True)

    def get_jobid(self, pool_id):
        return 'pypeliner_{}_{}'.format(pool_id, self.run_id)

    def cleanup_old_job_files(self, pickle_blob_path, pickle_local_path, temps_dir):
        # Delete any previous job result file locally and in blob
        self.batch_client.blob_client.delete_blob(
            container_name=self.batch_client.container_name, blob_name=pickle_blob_path
        )

        if os.path.exists(pickle_local_path):
            os.remove(pickle_local_path)

        # Delete any previous log files locally
        for debug_type, debug_filename in self.debug_filenames.items():
            debug_filename = os.path.join(temps_dir, debug_filename)
            if os.path.exists(debug_filename):
                os.remove(debug_filename)

    def create_run_script(self, command, script_path, pool_id):
        with open(script_path, 'w') as f:
            f.write('set -e\n')
            f.write('set -o pipefail\n\n')
            f.write(self.batch_client.compute_start_commands[pool_id] + '\n')
            f.write(command + '\n')
            f.write(self.batch_client.compute_finish_commands[pool_id] + '\n')
            f.write('wait')

    def send(self, ctx, name, sent, temps_dir):
        """ Add a job to the queue.

        Args:
            ctx (dict): context of job, mem etc.
            name (str): unique name for the job
            sent (callable): callable object
            temps_dir (str): unique path for strong job temps

        """
        pool_id = self.batch_client.find_pool(ctx)
        job_id = self.get_jobid(pool_id)
        task_id = self.batch_client.get_task_id(name)

        self.logger.debug(
            'assigning job {} to pool {}'.format(name, pool_id)
        )
        self.logger.info(
            'assigning task_id {} to job {}'.format(
                task_id,
                name))

        self.batch_client.create_pool(pool_id)
        self.batch_client.create_job(pool_id, job_id)

        self.mapping_from_tasks_to_job[task_id] = job_id
        self.job_names[task_id] = name
        self.job_task_ids[name] = task_id
        self.job_temps_dir[name] = temps_dir
        self.job_blobname_prefix[name] = 'output_' + task_id

        after_remote_path = 'job_result.pickle'
        after_local_path = os.path.join(temps_dir, after_remote_path)
        after_blob_path = os.path.join(self.job_blobname_prefix[name], after_remote_path)

        self.cleanup_old_job_files(after_blob_path, after_local_path, temps_dir)

        before_remote_path = 'job.pickle'
        before_local_path = os.path.join(temps_dir, before_remote_path)
        before_blob_path = os.path.join(self.job_blobname_prefix[name], before_remote_path)

        with open(before_local_path, 'wb') as before:
            pickle.dump(sent, before)

        before_file_resource = self.batch_client.create_blob_resource(
            before_local_path, before_blob_path, before_remote_path
        )

        # Create a bash script for running the job
        script_remote_path = 'job.sh'
        script_local_path = os.path.join(temps_dir, script_remote_path)
        script_blob_path = os.path.join(
            self.job_blobname_prefix[name], script_remote_path
        )

        command = ['pypeliner_delegate', '$AZ_BATCH_TASK_WORKING_DIR/' + before_remote_path,
                   '$AZ_BATCH_TASK_WORKING_DIR/' + after_remote_path]

        # local tasks like setobj need to run without docker args,
        # since they might already be running inside docker
        if not ctx.get('local'):
            command = pypeliner.containerize.containerize_args(*command)

        command = ' '.join(command)

        self.create_run_script(command, script_local_path, pool_id)

        run_script_file = self.batch_client.create_blob_resource(
            script_local_path, script_blob_path, script_remote_path)

        # Add the task to the job
        self.batch_client.add_task(
            job_id, task_id, before_file_resource,
            run_script_file, self.job_blobname_prefix[name]
        )

        self.running_task_ids.add(task_id)

    def wait(self, immediate=False):
        """ Wait for a job to finish.

        KwArgs:
            immediate (bool): do not wait if no job has finished

        Returns:
            str: job name

        """

        timeout = datetime.timedelta(minutes=60)

        while True:
            timeout_expiration = datetime.datetime.now() + timeout

            while datetime.datetime.now() < timeout_expiration:
                for task_id in self.running_task_ids.intersection(
                        self.completed_task_ids):
                    return self.job_names[task_id]

                if not self._update_task_state():
                    time.sleep(20)

            if not self.running_task_ids:
                self.logger.debug('something is not right. please try restarting the pipeline again')

            for task in self.running_task_ids:
                name = self.job_names[task]
                azurejob = self.mapping_from_tasks_to_job[task]
                self.logger.debug(
                    "waiting for task {} with name {} running under job {}".format(task, name, azurejob))
                self.batch_client.check_for_missing_node(azurejob, task)

            self.logger.warn(
                "Tasks did not reach 'Completed' state within timeout period of " +
                str(timeout))

            self.logger.info(
                "Most recent transition: {}".format(
                    self.most_recent_transition_time))
            task_filter = "state eq 'completed'"

            for pool_id in self.batch_client.active_pools:
                self.batch_client.check_pool_for_failed_nodes(pool_id)

            for job_id in self.batch_client.active_jobs:
                tasks = list(
                    self.batch_client.get_task_list(
                        job_id,
                        task_filter=task_filter)
                )
                self.logger.info("Received total {} tasks".format(len(tasks)))
                for task in tasks:
                    if task.id not in self.completed_task_ids:
                        self.logger.info(
                            "Missed completed task: {}".format(
                                task.serialize()))
                        self.completed_task_ids.add(task.id)

    def _update_task_state(self, latest_transition_time=None):
        """ Query azure and update task state. """

        list_max_results = 1000

        task_filter = "state eq 'completed'"

        if self.most_recent_transition_time is not None:
            assert self.most_recent_transition_time.tzname() == 'UTC'
            filter_time_string = self.most_recent_transition_time.strftime(
                "%Y-%m-%dT%H:%M:%SZ")
            task_filter += " and stateTransitionTime gt DateTime'{}'".format(
                filter_time_string)

        if latest_transition_time is not None:
            assert latest_transition_time.tzname() == 'UTC'
            filter_time_string = latest_transition_time.strftime(
                "%Y-%m-%dT%H:%M:%SZ")
            task_filter += " and stateTransitionTime lt DateTime'{}'".format(
                filter_time_string)

        tasks = []
        for job_id in self.batch_client.active_jobs:
            tasks += list(self.batch_client.get_task_list(
                job_id, task_filter=task_filter, list_max_results=list_max_results
            )
            )
        sorted_transition_times = sorted(
            [task.state_transition_time for task in tasks])

        if len(tasks) >= list_max_results:
            mid_transition_time = sorted_transition_times[
                int(len(sorted_transition_times) / 2)]
            return self._update_task_state(
                latest_transition_time=mid_transition_time)

        elif len(tasks) == 0:
            return False

        self.most_recent_transition_time = sorted_transition_times[-1]

        num_completed_before = len(self.completed_task_ids)
        self.completed_task_ids.update([task.id for task in tasks])
        return len(self.completed_task_ids) > num_completed_before

    def _create_error_text(self, job_temp_dir, hostname=None):
        error_text = 'failed to execute'

        if hostname:
            error_text += ' on node {}'.format(hostname)

        error_text = [error_text]

        for debug_type, debug_filename in self.debug_filenames.items():
            debug_filename = os.path.join(job_temp_dir, debug_filename)

            if not os.path.exists(debug_filename):
                error_text += [debug_type + ': missing']

                continue

            with open(debug_filename, 'r') as debug_file:
                error_text += [debug_type + ':']

                for line in debug_file:
                    error_text += ['\t' + line.rstrip()]

        return '\n'.join(error_text)

    def receive(self, name):
        """ Receive finished job.

        Args:
            name (str): name of job to retrieve

        Returns:
            object: received object

        """

        task_id = self.job_task_ids.pop(name)
        self.job_names.pop(task_id)
        temps_dir = self.job_temps_dir.pop(name)
        blobname_prefix = self.job_blobname_prefix.pop(name)
        self.running_task_ids.remove(task_id)

        job_id = self.mapping_from_tasks_to_job[task_id]

        # hostname on the node doesnt match id in azure ui(affinity id in API)
        # overrides the hostname collected from node with affinity id.
        hostname = self.batch_client.get_node_id_for_task(job_id, task_id)

        self.batch_client.download_blobs_from_container(
            temps_dir,
            blobname_prefix)

        job_after_filename = os.path.join(temps_dir, 'job_result.pickle')

        if not os.path.exists(job_after_filename):
            raise pypeliner.execqueue.base.ReceiveError(
                self._create_error_text(temps_dir, hostname=hostname))

        with open(job_after_filename, 'rb') as job_after_file:
            received = pickle.load(job_after_file)

        if received is None:
            raise pypeliner.execqueue.base.ReceiveError(
                self._create_error_text(temps_dir, hostname=hostname))

        # just in case a new storage key was pulled from storage in job
        # generally this doesnt happen as key gets pulled on head node to
        # get the create time for all resources anyway prior to running the job
        pypeliner.helpers.GlobalState.update_all(received.pypeliner_globals)

        received.hostname = hostname

        # batch keeps scheduling tasks on nodes in failed state, need to
        # explicitly disable failed nodes.
        if not received.finished or not received.started:
            pool_id = self.batch_client.get_pool_id_from_job(job_id)
            self.batch_client.verify_node_status(hostname, pool_id)

        return received

    @property
    def length(self):
        """ Number of jobs in the queue. """
        return len(self.job_names)

    @property
    def empty(self):
        """ Queue is empty. """
        return self.length == 0


if __name__ == "__main__":
    exc_dir = 'tempazure'

    logging.basicConfig(level=logging.DEBUG)

    with AzureJobQueue() as queue:
        job = pypeliner.tests.jobs.TestJob()
        pypeliner.helpers.makedirs(exc_dir)

        queue.send({}, 'testjob1', job, exc_dir)

        finished_name = queue.wait()

        job2 = queue.receive(finished_name)



import sqlite3


class SqliteDb(object):
    """
    Wrapper class for interacting with sqlite databases as a key value store
    :param filename: path at which the db file will be stored
    :type filename: any
    """

    def __init__(self, filename):
        self.conn = sqlite3.connect(filename)
        self.conn.execute("CREATE TABLE IF NOT EXISTS kv (key text unique, value text)")
        self.c = self.conn.cursor()

    def close(self):
        """
        close the database connections
        """
        self.conn.commit()
        self.conn.close()

    def __len__(self):
        self.c.execute('SELECT COUNT(*) FROM kv')
        rows = self.c.fetchone()[0]
        return rows if rows is not None else 0

    def iterkeys(self):
        """
        iterator over the keys in database
        :return: key value
        :rtype: any
        """
        c1 = self.conn.cursor()
        for row in c1.execute('SELECT key FROM kv'):
            yield row[0]

    def itervalues(self):
        """
        iterator over values in the database
        :return: value in database
        :rtype: any
        """
        c2 = self.conn.cursor()
        for row in c2.execute('SELECT value FROM kv'):
            yield row[0]

    def iteritems(self):
        """
        iterator over keys and values
        :return: tuple of key and its value
        :rtype: tuple
        """
        c3 = self.conn.cursor()
        for row in c3.execute('SELECT key, value FROM kv'):
            yield row[0], row[1]

    def keys(self):
        """
        returns all keys
        :return: list of keys
        :rtype: list
        """
        return [v for v in self.iterkeys()]

    def values(self):
        """
        returns all values
        :return: list of values
        :rtype: list
        """
        return [v for v in self.itervalues()]

    def items(self):
        """
        all keys and values
        :return: list of tuple of key and its value
        :rtype: list
        """
        return [v for v in self.iteritems()]

    def delete(self, key):
        c3 = self.conn.cursor()
        c3.execute('DELETE FROM kv WHERE key = ?', (key,))


    def __contains__(self, key):
        """
        check if a key exists in database
        """
        self.c.execute('SELECT 1 FROM kv WHERE key = ?', (key,))
        return self.c.fetchone() is not None

    def __getitem__(self, key):
        self.c.execute('SELECT value FROM kv WHERE key = ?', (key,))
        item = self.c.fetchone()
        if item is None:
            raise KeyError(key)
        return item[0]

    def get(self, key, default=None):
        """
        get the key if it exists, return default otherwise
        :param key: key to look for
        :type key: any
        :param default: value to return if key doesnt exist
        :type default: any
        :return: value for the key or default
        :rtype: any
        """
        try:
            return self.__getitem__(key)
        except KeyError:
            return default

    def __setitem__(self, key, value):
        self.c.execute('REPLACE INTO kv (key, value) VALUES (?,?)', (key, value))
        self.conn.commit()

    def __delitem__(self, key):
        if key not in self:
            raise KeyError(key)
        self.c.execute('DELETE FROM kv WHERE key = ?', (key,))
        self.conn.commit()

    def __iter__(self):
        return iter(self.items())



import sys
_pypeliner_internal_global_state = {}

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions

__all__ = ['helpers', 'scheduler', 'commandline', 'execqueue', 'delegator', 'app']

import pypeliner.helpers as helpers
import pypeliner.scheduler as scheduler
import pypeliner.execqueue as execqueue
import pypeliner.commandline as commandline
import pypeliner.delegator as delegator
import pypeliner.app as app



import logging

import pypeliner.identifiers
import dill as pickle


class Dependency(object):
    """ An input/output in the dependency graph
    
    This class is mainly used for tracking dependencies in the
    dependency graph for which it is not appropriate to check
    timestamps, in particular for an axis chunk input to a job
    parallelized on that axis.
    """
    is_temp = False

    def __init__(self, name, node):
        self.name = name
        self.node = node

    @property
    def id(self):
        return (self.name, self.node)

    @property
    def exists(self):
        return True

    def build_displayname(self, base_node=pypeliner.identifiers.Node()):
        name = '/' + self.name
        if self.node.displayname != '':
            name = '/' + self.node.displayname + name
        if base_node.displayname != '':
            name = '/' + base_node.displayname + name
        return name

    def cleanup(self):
        pass


class Resource(Dependency):
    """ Abstract input/output in the dependency graph
    associated with a file tracked using creation time """

    def build_displayname_filename(self, base_node=pypeliner.identifiers.Node()):
        displayname = self.build_displayname(base_node)
        if displayname != self.filename:
            return ' '.join([displayname, self.filename])
        else:
            return displayname

    @property
    def exists(self):
        raise NotImplementedError

    @property
    def createtime(self):
        raise NotImplementedError

    def touch(self):
        raise NotImplementedError

    def allocate(self):
        self.store.allocate()
        for store in self.extra_stores:
            store.allocate()

    def push(self):
        self.store.push()
        for store in self.extra_stores:
            store.push()

    def pull(self):
        self.store.pull()
        for store in self.extra_stores:
            store.pull()


class UserResource(Resource):
    """ A file resource with filename and creation time if created """

    def __init__(self, storage, name, node, filename, direct_write=False, extensions=None, store_dir=None):
        self.name = name
        self.node = node
        # Note:
        # filenames may be none if they are not yet defined
        # for example for a job that creates axis chunks that
        # are used by the same job's outputs
        self.extra_stores = []
        if filename is None:
            self.store = None
            self.filename = None
        else:
            self.store = storage.create_store(
                filename, is_temp=False, direct_write=direct_write, store_dir=store_dir
            )
            self.filename = self.store.filename
            if extensions is not None:
                for ext in extensions:
                    self.extra_stores.append(
                        storage.create_store(
                            filename, extension=ext, is_temp=False, direct_write=direct_write,
                            store_dir=store_dir
                        )
                    )

    def build_displayname(self, base_node=pypeliner.identifiers.Node()):
        return self.filename

    @property
    def exists(self):
        if self.store is None:
            return None
        return self.store.get_exists()

    @property
    def createtime(self):
        if self.store is None:
            return None
        return self.store.get_createtime()

    @property
    def write_filename(self):
        return self.store.write_filename

    def touch(self):
        if not self.exists:
            raise Exception('cannot touch missing user output')
        self.store.touch()


class TempFileResource(Resource):
    """ A file resource with filename and creation time if created """

    def __init__(self, storage, name, node, filename, direct_write=False, extensions=None, store_dir=None):
        self.name = name
        self.node = node
        self.is_temp = True
        self.store = storage.create_store(
            filename, is_temp=True, direct_write=direct_write, store_dir=store_dir
        )
        self.filename = self.store.filename
        self.extra_stores = []
        if extensions is not None:
            for ext in extensions:
                self.extra_stores.append(
                    storage.create_store(
                        filename, extension=ext, is_temp=False, direct_write=direct_write,
                        store_dir=store_dir
                    )
                )

    @property
    def exists(self):
        return self.store.get_exists()

    @property
    def createtime(self):
        return self.store.get_createtime()

    @property
    def write_filename(self):
        return self.store.write_filename

    def touch(self):
        self.store.touch()

    def cleanup(self):
        if self.exists:
            logging.getLogger('pypeliner.resources').debug('removing ' + self.filename)
            self.store.delete()


class TempObjResource(Resource):
    """ A file resource with filename and creation time if created """

    def __init__(self, storage, name, node, filename, is_input=True):
        self.name = name
        self.node = node
        self.filename = filename + ('._i', '._o')[is_input]
        self.is_input = is_input
        self.is_temp = True
        self.store = storage.create_store(self.filename, is_temp=True)

    @property
    def exists(self):
        return self.store.get_exists()

    @property
    def createtime(self):
        return self.store.get_createtime()

    def touch(self):
        if not self.exists:
            raise Exception('cannot touch missing object')
        self.store.touch()

    def get_obj(self):
        self.store.allocate()
        self.store.pull()

        with open(self.store.filename, 'rb') as f:
            return pickle.load(f)

    def put_obj(self, obj):
        self.store.allocate()
        with open(self.store.write_filename, 'wb') as f:
            pickle.dump(obj, f)
        self.store.push()


def obj_equal(obj1, obj2):
    try:
        equal = obj1.__eq__(obj2)
        if equal is not NotImplemented:
            return equal
    except AttributeError:
        pass
    if obj1.__class__ != obj2.__class__:
        return False
    try:
        return obj1.__dict__ == obj2.__dict__
    except AttributeError:
        pass
    return obj1 == obj2


class TempObjManager(object):
    """ A file resource with filename and creation time if created """

    def __init__(self, storage, name, node, filename):
        self.name = name
        self.node = node
        self.input = TempObjResource(storage, name, node, filename, is_input=True)
        self.output = TempObjResource(storage, name, node, filename, is_input=False)

    def get_obj(self):
        if self.input.exists:
            try:
                return self.input.get_obj()
            except:
                return

    def finalize(self, obj):
        self.output.put_obj(obj)

        existing_obj = self.get_obj()

        if existing_obj is None or not obj_equal(obj, existing_obj):
            self.input.put_obj(obj)



import os

import pypeliner.helpers
import pypeliner.identifiers
import pypeliner.resources


class Arg(object):
    is_split = False

    def get_inputs(self):
        return []

    def get_merge_inputs(self):
        return []

    def get_outputs(self):
        return []

    def get_split_outputs(self):
        return []

    def resolve(self):
        return None

    def update(self, job):
        pass

    def finalize(self, v):
        pass

    def allocate(self):
        pass

    def pull(self):
        pass

    def push(self):
        pass


class SplitMergeArg(object):
    def get_node_chunks(self, node):
        chunks = tuple([a[1] for a in node[-len(self.axes):]])
        if len(chunks) == 1:
            return chunks[0]
        return chunks

    def get_axes_origin(self, axes_origin):
        if axes_origin is None:
            return set(range(len(self.axes)))
        return axes_origin


class TemplateArg(Arg):
    """ Templated name argument

    The name parameter is treated as a string with named formatting.  Resolves to the name formatted using the node
    dictionary.

    """

    def __init__(self, db, name, node, template=None, **kwargs):
        self.name = name
        self.node = node
        if template is not None:
            self.filename = template.format(**dict(node))
        else:
            self.filename = name.format(**dict(node))

    def resolve(self):
        return self.filename


class TempSpaceArg(Arg):
    """ Temporary space argument

    Resolves to a filename/directory contained within the temporary files directory.

    """

    def __init__(self, db, name, node, cleanup='both', **kwargs):
        self.name = name
        self.node = node
        self.cleanup = cleanup
        filename = db.get_temp_filename(name, node)
        self.resource = pypeliner.resources.UserResource(
            db.file_storage, name, node, filename,
            direct_write=kwargs.get('direct_write'),
            store_dir=kwargs.get('store_dir'),
            extensions=kwargs.get('extensions'))
        self.filename = self.resource.filename

    def allocate(self):
        self.resource.allocate()

    def get_outputs(self):
        yield pypeliner.resources.Dependency(self.name, self.node)

    def resolve(self):
        if self.node.undefined:
            raise Exception('Undefined node {} for {}'.format(self.node.displayname, self.name))
        return self.filename

    def pull(self):
        if self.cleanup in ('before', 'both'):
            pypeliner.helpers.removefiledir(self.filename)
        pypeliner.helpers.makedirs(os.path.dirname(self.filename))

    def push(self):
        if self.cleanup in ('after', 'both'):
            pypeliner.helpers.removefiledir(self.filename)


class MergeTemplateArg(Arg, SplitMergeArg):
    """ Temp input files merged along a single axis

    The name parameter is treated as a string with named formatting.  Resolves to a dictionary with the keys as chunks
    for the merge axis.  Each value is the name formatted using the merge node dictionary.

    """

    def __init__(self, db, name, node, axes, template=None, **kwargs):
        self.name = name
        self.node = node
        self.axes = axes
        self.template = template
        self.merge_inputs = list(db.nodemgr.get_merge_inputs(self.axes, self.node))
        self.formatted = {}
        for node in db.nodemgr.retrieve_nodes(self.axes, self.node):
            if self.template is not None:
                self.formatted[self.get_node_chunks(node)] = self.template.format(**dict(node))
            else:
                self.formatted[self.get_node_chunks(node)] = self.name.format(**dict(node))

    def get_merge_inputs(self):
        return self.merge_inputs

    def resolve(self):
        return self.formatted


class InputFileArg(Arg):
    """ Input file argument

    The name argument is treated as a filename with named formatting.  Resolves to a filename formatted using the node
    dictionary.

    """

    def __init__(self, db, name, node, fnames=None, template=None, **kwargs):
        filename = db.get_user_filename(name, node, fnames=fnames, template=template)
        self.resource = pypeliner.resources.UserResource(
            db.file_storage, name, node, filename,
            direct_write=kwargs.get('direct_write'),
            store_dir=kwargs.get('store_dir'),
            extensions=kwargs.get('extensions'))

    def get_inputs(self):
        yield self.resource

    def resolve(self):
        return self.resource.filename

    def allocate(self):
        self.resource.allocate()

    def pull(self):
        self.resource.pull()


class MergeFileArg(Arg, SplitMergeArg):
    """ Input files merged along a single axis

    The name argument is treated as a filename with named formatting.  Resolves to a dictionary with keys as chunks for
    the merge axis.  Each value is the filename formatted using the merge node dictonary.

    """

    def __init__(self, db, name, node, axes, fnames=None, template=None, **kwargs):
        self.name = name
        self.node = node
        self.axes = axes
        self.fnames = fnames
        self.template = template
        self.resources = []
        for node in db.nodemgr.retrieve_nodes(self.axes, self.node):
            filename = db.get_user_filename(self.name, node, fnames=self.fnames, template=self.template)
            resource = pypeliner.resources.UserResource(
                db.file_storage, self.name, node, filename,
                direct_write=kwargs.get('direct_write'),
                store_dir=kwargs.get('store_dir'),
                extensions=kwargs.get('extensions'))
            self.resources.append(resource)
        self.merge_inputs = []
        for dependency in db.nodemgr.get_merge_inputs(self.axes, self.node):
            self.merge_inputs.append(dependency)

    def get_inputs(self):
        return self.resources

    def get_merge_inputs(self):
        return self.merge_inputs

    def resolve(self):
        resolved = dict()
        for resource in self.resources:
            resolved[self.get_node_chunks(resource.node)] = resource.filename
        return resolved

    def allocate(self):
        for resource in self.resources:
            resource.allocate()

    def pull(self):
        for resource in self.resources:
            resource.pull()


class OutputFileArg(Arg):
    """ Output file argument

    The name argument is treated as a filename with named formatting.  Resolves to a filename formatted using the node
    dictionary, including the '.tmp' suffix.

    """

    def __init__(self, db, name, node, fnames=None, template=None, **kwargs):
        filename = db.get_user_filename(name, node, fnames=fnames, template=template)
        self.resource = pypeliner.resources.UserResource(
            db.file_storage, name, node, filename,
            direct_write=kwargs.get('direct_write'),
            store_dir=kwargs.get('store_dir'),
            extensions=kwargs.get('extensions'))

    def get_outputs(self):
        yield self.resource

    def resolve(self):
        return self.resource.write_filename

    def allocate(self):
        self.resource.allocate()

    def push(self):
        self.resource.push()


class SplitFileArg(Arg, SplitMergeArg):
    """ Output file arguments from a split

    The name argument is treated as a filename with named formatting.  Resolves to a filename callback that can be used
    to generate filenames based on a given split axis chunk.  Resolved filenames have the '.tmp' suffix.  Finalizing
    involves removing the '.tmp' suffix for each file created by the job.

    """

    def __init__(self, db, name, node, axes, axes_origin=None, fnames=None, template=None, **kwargs):
        self.name = name
        self.node = node
        self.axes = axes
        self.axes_origin = self.get_axes_origin(axes_origin)
        self.is_split = len(self.axes_origin) != 0
        self.fnames = fnames
        self.template = template
        self.resources = []
        for node in db.nodemgr.retrieve_nodes(self.axes, self.node):
            filename = db.get_user_filename(self.name, node, fnames=self.fnames, template=self.template)
            resource = pypeliner.resources.UserResource(
                db.file_storage, self.name, node, filename,
                direct_write=kwargs.get('direct_write'),
                store_dir=kwargs.get('store_dir'),
                extensions=kwargs.get('extensions'))
            self.resources.append(resource)
        self.split_outputs = list(db.nodemgr.get_split_outputs(self.axes, self.node, subset=self.axes_origin))
        self.merge_inputs = list(db.nodemgr.get_merge_inputs(self.axes, self.node, subset=self.axes_origin))
        filename_creator = db.get_user_filename_creator(
            self.name, self.node.axes + self.axes, fnames=self.fnames, template=self.template)
        self.filename_callback = UserFilenameCallback(
            db.file_storage, self.name, self.node, self.axes, filename_creator, **kwargs)

    def get_merge_inputs(self):
        return self.merge_inputs

    def get_outputs(self):
        return self.resources

    def get_split_outputs(self):
        return self.split_outputs

    def resolve(self):
        return self.filename_callback

    def update(self, job):
        if self.is_split:
            job.workflow.db.nodemgr.store_chunks(self.axes, self.node, list(self.filename_callback.resources.keys()), subset=self.axes_origin)

    def push(self):
        for resource in self.filename_callback.resources.values():
            resource.push()


class TempInputObjArg(Arg):
    """ Temporary input object argument

    Resolves to an object.  If func is given, resolves to the return value of func called with object as the only
    parameter.

    """

    def __init__(self, db, name, node, func=None, **kwargs):
        filename = db.get_temp_filename(name, node)
        self.resource = pypeliner.resources.TempObjManager(db.file_storage, name, node, filename)
        self.func = func

    def get_inputs(self):
        yield self.resource.input

    def resolve(self):
        obj = self.resource.get_obj()
        if self.func is not None:
            obj = self.func(obj)
        return obj


class TempMergeObjArg(Arg, SplitMergeArg):
    """ Temp input object arguments merged along single axis

    Resolves to an dictionary of objects with keys given by the merge axis chunks.

    """

    def __init__(self, db, name, node, axes, func=None, **kwargs):
        self.name = name
        self.node = node
        self.axes = axes
        self.func = func
        self.resources = []
        for node in db.nodemgr.retrieve_nodes(self.axes, self.node):
            filename = db.get_temp_filename(self.name, node)
            resource = pypeliner.resources.TempObjManager(db.file_storage, self.name, node, filename)
            self.resources.append(resource)
        self.merge_inputs = list(db.nodemgr.get_merge_inputs(self.axes, self.node))

    def get_inputs(self):
        for resource in self.resources:
            yield resource.input

    def get_merge_inputs(self):
        return self.merge_inputs

    def resolve(self):
        resolved = dict()
        for resource in self.resources:
            obj = resource.get_obj()
            if self.func is not None:
                obj = self.func(obj)
            resolved[self.get_node_chunks(resource.node)] = obj
        return resolved


class TempOutputObjArg(Arg):
    """ Temporary output object argument

    Stores an object created by a job.

    """

    def __init__(self, db, name, node, **kwargs):
        filename = db.get_temp_filename(name, node)
        self.resource = pypeliner.resources.TempObjManager(db.file_storage, name, node, filename)

    def get_outputs(self):
        yield self.resource.output

    def resolve(self):
        return self

    def finalize(self, value):
        self.resource.finalize(value)


class TempSplitObjArg(Arg, SplitMergeArg):
    """ Temporary output object arguments from a split

    Stores a dictionary of objects created by a job.  The keys of the dictionary are taken as the chunks for the given
    split axis.

    """

    def __init__(self, db, name, node, axes, axes_origin=None, **kwargs):
        self.name = name
        self.node = node
        self.axes = axes
        self.axes_origin = self.get_axes_origin(axes_origin)
        self.is_split = len(self.axes_origin) != 0
        self.resources = []
        for node in db.nodemgr.retrieve_nodes(self.axes, self.node):
            filename = db.get_temp_filename(self.name, node)
            resource = pypeliner.resources.TempObjManager(db.file_storage, self.name, node, filename)
            self.resources.append(resource)
        self.merge_inputs = list(db.nodemgr.get_merge_inputs(self.axes, self.node, subset=self.axes_origin))
        self.split_outputs = list(db.nodemgr.get_split_outputs(self.axes, self.node, subset=self.axes_origin))
        self.storage = db.file_storage
        self.filename_creator = db.get_temp_filename_creator()

    def get_merge_inputs(self):
        return self.merge_inputs

    def get_outputs(self):
        for resource in self.resources:
            yield resource.output

    def get_split_outputs(self):
        return self.split_outputs

    def resolve(self):
        return self

    def finalize(self, values):
        self.chunks_list = list(values.keys())
        self.resources = []
        for chunks, value in values.items():
            if not isinstance(chunks, tuple):
                chunks = (chunks,)
            if len(self.axes) != len(chunks):
                raise ValueError('expected ' + str(len(self.axes)) + ' values for axes ' + str(self.axes))
            node = self.node
            for axis, chunk in zip(self.axes, chunks):
                node += pypeliner.identifiers.AxisInstance(axis, chunk)
            filename = self.filename_creator(self.name, node)
            resource = pypeliner.resources.TempObjManager(self.storage, self.name, node, filename)
            resource.finalize(value)
            self.resources.append(resource)

    def update(self, job):
        job.workflow.db.nodemgr.store_chunks(self.axes, self.node, self.chunks_list, subset=self.axes_origin)


class TempInputFileArg(Arg):
    """ Temp input file argument

    Resolves to a filename for a temporary file.

    """

    def __init__(self, db, name, node, **kwargs):
        filename = db.get_temp_filename(name, node)
        self.resource = pypeliner.resources.TempFileResource(
            db.file_storage, name, node, filename,
            direct_write=kwargs.get('direct_write'),
            store_dir=kwargs.get('store_dir'),
            extensions=kwargs.get('extensions'))

    def get_inputs(self):
        yield self.resource

    def resolve(self):
        return self.resource.filename

    def allocate(self):
        self.resource.allocate()

    def pull(self):
        self.resource.pull()


class TempMergeFileArg(Arg, SplitMergeArg):
    """ Temp input files merged along a single axis

    Resolves to a dictionary of filenames of temporary files.

    """

    def __init__(self, db, name, node, axes, **kwargs):
        self.name = name
        self.node = node
        self.axes = axes
        self.resources = []
        for node in db.nodemgr.retrieve_nodes(self.axes, self.node):
            filename = db.get_temp_filename(self.name, node)
            resource = pypeliner.resources.TempFileResource(
                db.file_storage, self.name, node, filename,
                direct_write=kwargs.get('direct_write'),
                store_dir=kwargs.get('store_dir'),
                extensions=kwargs.get('extensions'))
            self.resources.append(resource)
        self.merge_inputs = list(db.nodemgr.get_merge_inputs(self.axes, self.node))

    def get_inputs(self):
        return self.resources

    def get_merge_inputs(self):
        return self.merge_inputs

    def resolve(self):
        resolved = dict()
        for resource in self.resources:
            resolved[self.get_node_chunks(resource.node)] = resource.filename
        return resolved

    def allocate(self):
        for resource in self.resources:
            resource.allocate()

    def pull(self):
        for resource in self.resources:
            resource.pull()


class TempOutputFileArg(Arg):
    """ Temp output file argument

    Resolves to an output filename for a temporary file.  Finalizes with resource manager.

    """

    def __init__(self, db, name, node, **kwargs):
        filename = db.get_temp_filename(name, node)
        self.resource = pypeliner.resources.TempFileResource(
            db.file_storage, name, node, filename,
            direct_write=kwargs.get('direct_write'),
            store_dir=kwargs.get('store_dir'),
            extensions=kwargs.get('extensions'))

    def get_outputs(self):
        yield self.resource

    def resolve(self):
        return self.resource.write_filename

    def allocate(self):
        self.resource.allocate()

    def push(self):
        self.resource.push()


class FilenameCallback(object):
    """ Argument to split jobs providing callback for filenames
    with a particular instance """

    def __init__(self, storage, name, node, axes, filename_creator, **kwargs):
        self.storage = storage
        self.name = name
        self.node = node
        self.axes = axes
        self.filename_creator = filename_creator
        self.resources = dict()
        self.kwargs = kwargs

    def get_filename(self, *chunks):
        if len(self.axes) != len(chunks):
            raise ValueError('expected ' + str(len(self.axes)) + ' values for axes ' + str(self.axes))
        node = self.node
        for axis, chunk in zip(self.axes, chunks):
            node += pypeliner.identifiers.AxisInstance(axis, chunk)
        filename = self.filename_creator(self.name, node)
        resource = self.create_resource(
            self.storage, self.name, node, filename,
            direct_write=self.kwargs.get('direct_write'),
            store_dir=self.kwargs.get('store_dir'),
            extensions=self.kwargs.get('extensions'))
        return resource.write_filename

    def __call__(self, *chunks):
        return self.__getitem__(*chunks)

    def __getitem__(self, *chunks):
        if len(self.axes) != len(chunks):
            raise ValueError('expected ' + str(len(self.axes)) + ' values for axes ' + str(self.axes))
        node = self.node
        for axis, chunk in zip(self.axes, chunks):
            node += pypeliner.identifiers.AxisInstance(axis, chunk)
        filename = self.filename_creator(self.name, node)
        resource = self.create_resource(
            self.storage, self.name, node, filename,
            direct_write=self.kwargs.get('direct_write'),
            store_dir=self.kwargs.get('store_dir'),
            extensions=self.kwargs.get('extensions'))
        if len(chunks) == 1:
            self.resources[chunks[0]] = resource
        else:
            self.resources[chunks] = resource
        resource.allocate()
        return resource.write_filename

    def __repr__(self):
        return '{0}.{1}({2},{3},{4},{5})'.format(
            FilenameCallback.__module__,
            FilenameCallback.__name__,
            self.name,
            self.node,
            self.axes,
            self.filename_creator)


class TempFilenameCallback(FilenameCallback):
    create_resource = pypeliner.resources.TempFileResource


class UserFilenameCallback(FilenameCallback):
    create_resource = pypeliner.resources.UserResource


class TempSplitFileArg(Arg, SplitMergeArg):
    """ Temp output file arguments from a split

    Resolves to a filename callback that can be used to create a temporary filename for each chunk of the split on the
    given axis.  Finalizes with resource manager to move from temporary filename to final filename.

    """

    def __init__(self, db, name, node, axes, axes_origin=None, **kwargs):
        self.name = name
        self.node = node
        self.axes = axes
        self.axes_origin = self.get_axes_origin(axes_origin)
        self.is_split = len(self.axes_origin) != 0
        self.resources = []
        for node in db.nodemgr.retrieve_nodes(self.axes, self.node):
            filename = db.get_temp_filename(self.name, node)
            resource = pypeliner.resources.TempFileResource(
                db.file_storage, self.name, node, filename,
                direct_write=kwargs.get('direct_write'),
                store_dir=kwargs.get('store_dir'),
                extensions=kwargs.get('extensions'))
            self.resources.append(resource)
        self.merge_inputs = list(db.nodemgr.get_merge_inputs(self.axes, self.node, subset=self.axes_origin))
        self.split_outputs = list(db.nodemgr.get_split_outputs(self.axes, self.node, subset=self.axes_origin))
        self.filename_callback = TempFilenameCallback(
            db.file_storage, self.name, self.node, self.axes,
            db.get_temp_filename_creator(), **kwargs)

    def get_merge_inputs(self):
        return self.merge_inputs

    def get_outputs(self):
        return self.resources

    def get_split_outputs(self):
        return self.split_outputs

    def resolve(self):
        return self.filename_callback

    def update(self, job):
        if self.is_split:
            job.workflow.db.nodemgr.store_chunks(self.axes, self.node, list(self.filename_callback.resources.keys()), subset=self.axes_origin)

    def push(self):
        for resource in self.filename_callback.resources.values():
            resource.push()


class InputInstanceArg(Arg):
    """ Instance of a job as an argument

    Resolves to the instance of the given job for a specific axis.

    """

    def __init__(self, db, node, axis, **kwargs):
        self.chunk = dict(node)[axis]

    def resolve(self):
        return self.chunk


class InputChunksArg(Arg):
    """ Instance list of an axis as an argument

    Resolves to the list of chunks for the given axes.

    """

    def __init__(self, db, name, node, axis, **kwargs):
        self.node = node
        self.axis = axis
        self.merge_inputs = list(db.nodemgr.get_merge_inputs(self.axis, self.node))
        self.chunks = list(db.nodemgr.retrieve_chunks(self.axis, self.node))
        if len(self.axis) == 1:
            self.chunks = [chunk[0] for chunk in self.chunks]

    def get_merge_inputs(self):
        return self.merge_inputs

    def resolve(self):
        return self.chunks


class OutputChunksArg(Arg, SplitMergeArg):
    """ Instance list of a job as an argument

    Sets the list of chunks for the given axes.

    """

    def __init__(self, db, name, node, axes, axes_origin=None, **kwargs):
        self.node = node
        self.axes = axes
        self.axes_origin = self.get_axes_origin(axes_origin)
        self.is_split = len(self.axes_origin) != 0
        self.merge_inputs = list(db.nodemgr.get_merge_inputs(self.axes, self.node, subset=self.axes_origin))
        self.split_outputs = list(db.nodemgr.get_split_outputs(self.axes, self.node, subset=self.axes_origin))

    def get_merge_inputs(self):
        return self.merge_inputs

    def get_split_outputs(self):
        return self.split_outputs

    def resolve(self):
        return self

    def finalize(self, value):
        self.chunks_list = value

    def update(self, job):
        job.workflow.db.nodemgr.store_chunks(self.axes, self.node, self.chunks_list, subset=self.axes_origin)


class OutputWorkflowArg(Arg):
    """ Output workflow object argument

    Stores a workflow object created by a job.

    """

    def __init__(self, db, name, node, **kwargs):
        filename = db.get_temp_filename(name, node)
        self.resource = pypeliner.resources.TempObjManager(db.file_storage, name, node, filename)

    def get_outputs(self):
        yield self.resource.output

    def resolve(self):
        return self

    def finalize(self, value):
        self.resource.finalize(value)

    def get_obj(self):
        return self.resource.get_obj()





class JobQueue(object):
    """ Abstract class for a queue of jobs.
    """
    def send(self, ctx, name, sent, temps_dir):
        """ Add a job to the queue.

        Args:
            ctx (dict): context of job, mem etc.
            name (str): unique name for the job
            sent (callable): callable object
            temps_dir (str): unique path for strong job temps

        """
        raise NotImplementedError()

    def wait(self, immediate=False):
        """ Wait for a job to finish.

        KwArgs:
            immediate (bool): do not wait if no job has finished

        Returns:
            str: job name

        """
        raise NotImplementedError()

    def receive(self, name):
        """ Receive finished job.

        Args:
            name (str): name of job to retrieve

        Returns:
            object: received object

        """
        raise NotImplementedError()

    @property
    def length(self):
        """ Number of jobs in the queue. """
        raise NotImplementedError()

    @property
    def empty(self):
        """ Queue is empty. """
        raise NotImplementedError()


class SubmitError(Exception):
    pass


class ReceiveError(Exception):
    pass





"""
Job scheduling class

"""

import logging
import os
import traceback

import pypeliner
import pypeliner.database
import pypeliner.execqueue.base
import pypeliner.graph


class PipelineException(Exception):
    pass


class JobIdMismatchError(Exception):
    pass


class IncompleteJobException(Exception):
    pass


class Scheduler(object):
    """ Job scheduling class for queueing a set of jobs and running
    those jobs according to their dependencies.

    """

    def __init__(self):
        self._logger = logging.getLogger('pypeliner.scheduler')
        self.max_jobs = 1
        self.cleanup = True
        self.temps_dir = './tmp'
        self.workflow_dir = './'
        self.logs_dir = './log'
        self.freeze = True

    def __setattr__(self, attr, value):
        if not attr.startswith('_') and getattr(self, "freeze", False) and not hasattr(self, attr):
            raise AttributeError("Setting new attribute")
        super(Scheduler, self).__setattr__(attr, value)

    @property
    def workflow_dir(self):
        return self._workflow_dir

    @workflow_dir.setter
    def workflow_dir(self, value):
        self._workflow_dir = value

    @property
    def logs_dir(self):
        return self._logs_dir

    @logs_dir.setter
    def logs_dir(self, value):
        self._logs_dir = value

    def run(self, workflow_def, exec_queue, file_storage, runskip):
        """ Run the pipeline

        :param workflow_def: workflow of jobs to be submitted.
        :param exec_queue: queue to which jobs will be submitted.  The queues implemented
                           in :py:mod:`pypeliner.execqueue` should suffice for most purposes
        :param runskip: callable object returning boolean, used to determine whether to run jobs

        Call this function after adding jobs to a workflow using
        :py:func:`pypeliner.scheduler.Scheduler.transform` etc.  Jobs will be run locally or
        remotely using the `exec_queue` provided until completion.  On failure, the function
        will wait for the remaining jobs to finish but will not submit new ones.  The first
        interrupt (control-C) in this function will result in the sessation of new job creation,
        and the second interrupt will attempt to cleanly cancel all jobs.

        """

        self._active_jobs = dict()
        self._job_exc_dirs = set()
        with pypeliner.database.WorkflowDatabaseFactory(
                self.temps_dir, self.workflow_dir, self.logs_dir, file_storage
        ) as db_factory:
            workflow = pypeliner.graph.WorkflowInstance(
                workflow_def, db_factory, runskip, ctx=workflow_def.ctx, cleanup=self.cleanup
            )
            failing = False
            try:
                try:
                    while True:
                        self._add_jobs(exec_queue, workflow)
                        if exec_queue.empty:
                            break
                        self._wait_next_job(exec_queue, workflow)
                except KeyboardInterrupt as e:
                    raise
                except Exception:
                    failing = True
                    self._logger.error('exception\n' + traceback.format_exc())
                while not exec_queue.empty:
                    try:
                        self._wait_next_job(exec_queue, workflow)
                    except KeyboardInterrupt as e:
                        raise
                    except Exception:
                        self._logger.error('exception\n' + traceback.format_exc())
            except KeyboardInterrupt as e:
                self._logger.error('interrupted')
                raise
            if failing:
                self._logger.error('pipeline failed')
                raise PipelineException('pipeline failed')

    def _add_job(self, exec_queue, job):
        sent = job.create_callable()
        exc_dir = job.create_exc_dir()

        self._active_jobs[job.displayname] = job

        if exc_dir in self._job_exc_dirs:
            raise ValueError('duplicate temps directory ' + exc_dir)
        self._job_exc_dirs.add(exc_dir)

        exec_log = [('job_name', job.displayname), ('status', ' executing'), ('context', job.ctx)]

        if hasattr(job.job_def, 'sandbox'):
            if job.job_def.sandbox is not None:
                sandbox_name = os.path.basename(job.job_def.sandbox.prefix)
            else:
                sandbox_name = 'root'
            exec_log.append(('sandbox', sandbox_name))

        pypeliner.helpers.log_event(
            exec_log,
            extras={'task_name': job.id[1], "cmd": sent.displaycommand},
            logger=self._logger
        )

        dirs = [self.temps_dir, self.logs_dir]
        # need these to track the output file and the inputs
        dirs.extend([v.filename for v in job.inputs if isinstance(v, pypeliner.resources.UserResource)])
        dirs.extend([v.filename for v in job.outputs if isinstance(v, pypeliner.resources.UserResource)])
        sent.dirs = dirs
        sent.version = pypeliner.__version__

        exec_queue.send(job.ctx, job.displayname, sent, exc_dir)

    def _retry_job(self, exec_queue, job):
        if not job.retry():
            return False
        pypeliner.helpers.log_event(
            [('job_name', job.displayname), ('retry_count', job.retry_idx)],
            extras={'task_name': job.id[1]},
            logger=self._logger
        )
        self._add_job(exec_queue, job)
        return True

    def _add_jobs(self, exec_queue, workflow):
        while exec_queue.length < self.max_jobs:
            try:
                job = workflow.pop_next_job()
            except pypeliner.graph.NoJobs:
                return
            self._add_job(exec_queue, job)

    def _handle_error(self, job, error, error_str, exec_queue):
        pypeliner.helpers.log_event(
            ['job_name', job.displayname, error_str, error],
            extras={'job_name': job.displayname, 'task_name': job.id[1], "status": 'error'},
            logger=self._logger,
            level='err',
        )
        if not self._retry_job(exec_queue, job):
            raise pypeliner.graph.IncompleteJobException()


    def _wait_next_job(self, exec_queue, workflow):

        name = exec_queue.wait()

        job = self._active_jobs[name]
        del self._active_jobs[name]

        assert job is not None

        try:
            received = exec_queue.receive(name)
            received.collect_logs()
            if not received.finished:
                raise IncompleteJobException()
        except pypeliner.execqueue.base.ReceiveError:
            self._handle_error(job, traceback.format_exc(), "submit error\n", exec_queue)
            return
        except IncompleteJobException:
            self._handle_error(job, received.log_text(), "failed to complete\n", exec_queue)
            pypeliner.helpers.log_event(
                [('job_name', job.displayname), ('->', received.displaycommand)],
                extras={'task_name': job.id[1]},
                logger=self._logger
            )
            return
        except Exception:
            self._handle_error(job, traceback.format_exc(), 'collect logs error\n', exec_queue)
            return

        if job.id != received.id:
            raise JobIdMismatchError('job id {} doesnt match received id {}'.format(job.id, received.id))

        pypeliner.helpers.log_event(
            ['job_name', job.displayname, "completed successfully"],
            extras={'job_name': job.displayname, 'task_name': job.id[1], "status": 'success'},
            logger=self._logger
        )

        pypeliner.helpers.log_event(
            [('job_name', job.displayname), ('->', received.displaycommand)],
            extras={'task_name': job.id[1]},
            logger=self._logger
        )

        pypeliner.helpers.log_event(
            [('job_name', job.displayname), ('time', str(received.duration) + 's')],
            extras={'task_name': job.id[1]},
            logger=self._logger
        )

        pypeliner.helpers.log_event(
            [('job_name', job.displayname), ('memory', str(received.memoryused) + 'G')],
            extras={'task_name': job.id[1]},
            logger=self._logger
        )

        pypeliner.helpers.log_event(
            [('job_name', job.displayname), ('host_name', received.hostname)],
            extras={'task_name': job.id[1]},
            logger=self._logger
        )

        received.finalize(job)



import os

import pypeliner


def containerize_args(*args):
    context_cfg = pypeliner.helpers.GlobalState.get("context_config")

    if not context_cfg:
        return args
    elif context_cfg.get('singularity'):
        return singularity_args(args, context_cfg['singularity'])
    elif context_cfg.get('docker'):
        return dockerize_args(args, context_cfg['docker'])
    else:
        return args


def singularity_args(args, context_cfg):
    singularity = context_cfg.get('singularity_exe', 'singularity')
    image = context_cfg['image']

    if not image.endswith('.sif'):
        raise Exception('only sif images are allowed')

    singularity_command = [singularity, 'run']

    mounts = sorted(set(context_cfg.get("mounts", {}).values()))
    for mount in mounts:
        singularity_command.extend(['--bind', mount])

    wdir = os.getcwd()
    singularity_command.extend(['--pwd', wdir])
    singularity_command.append(image)

    singularity_command.extend(args)

    return singularity_command


def dockerize_args(args, context_cfg):
    docker_args = [
        'docker', 'run', '-w', '$PWD', '-v', '$PWD:$PWD',
        '--rm'
    ]

    mounts = sorted(set(context_cfg.get("mounts", {}).values()))
    for mount in mounts:
        docker_args.extend(['-v', '{}:{}'.format(mount, mount)])

    docker_args.append(context_cfg['image'])

    args = docker_args + list(args)

    return args



import os
import resource
import signal
import sys
import traceback

import datetime
import pypeliner.arguments
import pypeliner.deep
import pypeliner.helpers
import pypeliner.identifiers
import pypeliner.managed
import pypeliner.resources
import pypeliner.storage
import six
import socket
import time
import uuid
from datetime import timedelta
from collections import deque

import pypeliner.commandline

try:
    from future_builtins import filter
except ImportError:
    pass


class IncompleteWorkflowException(Exception):
    pass


class CallSet(object):
    """ Set of positional and keyword arguments, and a return value """

    def __init__(self, ret=None, args=None, kwargs=None):
        if ret is not None and not isinstance(ret, pypeliner.managed.Managed):
            raise ValueError('ret must be a managed object')
        self.ret = ret
        if args is None:
            self.args = ()
        elif not isinstance(args, (tuple, list)):
            raise ValueError('args must be a list or tuple')
        else:
            self.args = args
        if kwargs is None:
            self.kwargs = {}
        elif not isinstance(kwargs, dict):
            raise ValueError('kwargs must be a dict')
        else:
            self.kwargs = kwargs


class JobDefinition(object):
    """ Represents an abstract job including function and arguments """

    def __init__(self, name, axes, ctx, func, argset, sandbox=None):
        self.name = name
        self.axes = axes
        self.ctx = ctx
        self.func = func
        self.argset = argset
        self.sandbox = sandbox

    @property
    def wrapped_func(self):
        if self.sandbox is not None:
            return self.sandbox.wrap_function(self.func)
        else:
            return self.func

    def create_job_instances(self, workflow, db):
        for node in db.nodemgr.retrieve_nodes(self.axes):
            yield JobInstance(self, workflow, db, node)


def _pretty_date(ts):
    if ts is None:
        return 'none'
    return datetime.datetime.fromtimestamp(ts).strftime('%Y/%m/%d-%H:%M:%S')


class JobInstance(object):
    """ Represents a job including function and arguments """
    direct_write = False

    def __init__(self, job_def, workflow, db, node):
        self.job_def = job_def
        self.workflow = workflow
        self.db = db
        self.node = node
        temps_dir = os.path.join(db.temps_dir, self.node.subdir, self.job_def.name)
        self.store_dir = os.path.join(temps_dir, str(uuid.uuid1()))
        self.arglist = list()
        try:
            self.argset = pypeliner.deep.deeptransform(
                self.job_def.argset,
                self._create_arg)
        except pypeliner.managed.JobArgMismatchException as e:
            e.job_name = self.displayname
            raise
        self.logs_dir = os.path.join(db.logs_dir, self.node.subdir, self.job_def.name)
        self.retry_idx = 0
        self.ctx = job_def.ctx.copy()
        self.is_required_downstream = False
        self.init_inputs_outputs()
        self.runskip_request = None

    def _create_arg(self, mg):
        if not isinstance(mg, pypeliner.managed.Managed):
            return None, False
        arg = mg.create_arg(self)
        self.arglist.append(arg)
        return arg, True

    def init_inputs_outputs(self):
        self.inputs = list()
        self.outputs = list()
        merge_inputs = list()
        split_outputs = list()
        for arg in self.arglist:
            if isinstance(arg, pypeliner.arguments.Arg):
                self.inputs.extend(arg.get_inputs())
                self.outputs.extend(arg.get_outputs())
                merge_inputs.extend(arg.get_merge_inputs())
                split_outputs.extend(arg.get_split_outputs())
        # A dependency that is both a merge input and split output
        # is only an output
        split_output_ids = set([a.id for a in split_outputs])
        for input_ in merge_inputs:
            if input_.id not in split_output_ids:
                self.inputs.append(input_)
        self.outputs.extend(split_outputs)
        for node_input in self.db.nodemgr.get_node_inputs(self.node):
            self.inputs.append(node_input)

    @property
    def id(self):
        return (self.node, self.job_def.name)

    @property
    def jobname(self):
        return self.job_def.name

    @property
    def displayname(self):
        name = '/' + self.job_def.name
        if self.node.displayname != '':
            name = '/' + self.node.displayname + name
        if self.workflow.node.displayname != '':
            name = '/' + self.workflow.node.displayname + name
        return name

    @property
    def input_resources(self):
        return filter(lambda a: isinstance(a, pypeliner.resources.Resource), self.inputs)

    @property
    def output_resources(self):
        return filter(lambda a: isinstance(a, pypeliner.resources.Resource), self.outputs)

    def already_run(self):
        return self.db.job_shelf.get(self.displayname, False)

    def out_of_date(self):
        input_dates = [input.createtime for input in self.input_resources]
        output_dates = [output.createtime for output in self.output_resources]
        if len(input_dates) == 0 or len(output_dates) == 0:
            return True
        if None in output_dates:
            return True

        if None in input_dates:
            return True

        return max(input_dates) > min(output_dates)

    def explain_out_of_date(self):
        input_dates = [input.createtime for input in self.input_resources]
        output_dates = [output.createtime for output in self.output_resources]
        try:
            newest_input_date = max(input_dates)
        except (TypeError, ValueError):
            newest_input_date = None
        try:
            oldest_output_date = min(output_dates)
        except (TypeError, ValueError):
            oldest_output_date = None
        if len(input_dates) == 0 and len(output_dates) == 0:
            explanation = ['no inputs/outputs: always run']
        elif len(input_dates) == 0:
            explanation = ['no inputs: always run']
        elif len(output_dates) == 0:
            explanation = ['no outputs: always run']
        elif None in output_dates:
            explanation = ['missing outputs']
        elif newest_input_date > oldest_output_date:
            explanation = ['out of date outputs']
        elif self.is_required_downstream:
            explanation = ['outputs required by downstream job']
        else:
            explanation = ['up to date']
        for input in self.input_resources:
            status = ''
            if oldest_output_date is not None and input.createtime > oldest_output_date:
                status = 'new'
            text = 'input {0} {1} {2}'.format(
                input.build_displayname_filename(self.workflow.node),
                _pretty_date(input.createtime),
                status)
            explanation.append(text)
        for output in self.output_resources:
            if output.createtime is None:
                status = ''
            elif newest_input_date is not None and output.createtime < newest_input_date:
                status = '{0} old'.format(_pretty_date(output.createtime))
            else:
                status = '{0}'.format(_pretty_date(output.createtime))
            if not output.exists:
                status += ' missing'
            text = 'output {0} {1}'.format(
                output.build_displayname_filename(self.workflow.node),
                status)
            explanation.append(text)
        return '\n'.join(explanation)

    def output_missing(self):
        return not all([output.exists for output in self.output_resources])

    def touch_outputs(self):
        for output in self.output_resources:
            output.touch()

    def check_require_regenerate(self):
        for arg in self.arglist:
            if isinstance(arg, pypeliner.arguments.Arg):
                if arg.is_split:
                    return True
        return False

    def create_callable(self):
        return JobCallable(
            self.id, self.job_def.wrapped_func, self.argset, self.arglist,
            self.db.file_storage, self.store_dir, self.logs_dir, self.ctx
        )

    def create_exc_dir(self):
        exc_dir = os.path.join(self.logs_dir, 'exc{}'.format(self.retry_idx))
        pypeliner.helpers.makedirs(exc_dir)
        return exc_dir

    def update_ctx_value(self, retry_val, val, by_factor=False):
        if isinstance(val, (int, float, complex)):
            if by_factor:
                return val * retry_val
            else:
                return val + retry_val
        elif isinstance(val, six.string_types) and ':' in val:
            val = [int(i) for i in val.split(':')]
            if by_factor:
                assert isinstance(retry_val, (int, float)), \
                    "retry_factor for time should be an integer or float"
                val = [v * retry_val for v in val]
            else:
                retry_val = [int(i) for i in retry_val.split(':')]
                assert len(val) == len(retry_val)
                val = [v1 + v2 for v1, v2 in zip(val, retry_val)]
            val = [str(v).zfill(2) for v in val]
            return ':'.join(val)
        else:
            raise Exception('unknown_format')

    def retry(self):
        if self.retry_idx >= self.ctx.get('num_retry', 0):
            return False
        self.retry_idx += 1
        updated = False
        for key, value in self.ctx.items():
            if key.endswith('_retry_factor'):
                self.ctx[key[:-len('_retry_factor')]] = self.update_ctx_value(
                    value, self.ctx[key[:-len('_retry_factor')]], by_factor=True)
                updated = True
            elif key.endswith('_retry_increment'):
                self.ctx[key[:-len('_retry_increment')]] = self.update_ctx_value(
                    value, self.ctx[key[:-len('_retry_increment')]])
                updated = True
        return updated


class JobTimer(object):
    """ Timer using a context manager """

    def __init__(self):
        self._start = None
        self._finish = None

    def __enter__(self):
        self._start = time.time()

    def __exit__(self, exc_type, exc_value, traceback):
        self._finish = time.time()

    @property
    def duration(self):
        if self._finish is None or self._start is None:
            return '?'
        return int(self._finish - self._start)


class TimeOutError(Exception):
    """Exception Type for throwing timeout errors"""
    pass


class JobTimeOut(object):
    """ TimeOut using a context manager
        set an alarm for the timeout, catch it and throw an Exception.
    """

    def __init__(self, timeout):
        self._timeout_string = timeout

        if not timeout:
            self._timeout = None
            return

        if timeout.endswith("s"):
            multiplier = 1
        elif timeout.endswith("m"):
            multiplier = 60
        elif timeout.endswith("h"):
            multiplier = 60 * 60
        elif timeout.endswith("d"):
            multiplier = 60 * 60 * 24
        else:
            raise ValueError("Invalid timeout interval: {}." \
                             " Timeout should be a number followed by a suffix. " \
                             "SUFFIX may be s for seconds, m for minutes, h for hours " \
                             "or d for days".format(timeout))

        time_interval = int(timeout[:-1])

        self._timeout = time_interval * multiplier

    def handler(self, signum, frame):
        raise TimeOutError("Execution time exceeded the specified timeout of {}.".format(self._timeout_string))

    def __enter__(self):
        if self._timeout:
            signal.signal(signal.SIGALRM, self.handler)
            signal.alarm(self._timeout)

    def __exit__(self, exc_type, exc_value, traceback):
        pass


class JobMemoryTracker(object):
    """ track memory use"""

    def __init__(self):
        self._start = None
        self._finish = None

    def __enter__(self):
        self._start = 0

    def __exit__(self, exc_type, exc_value, traceback):
        # get memory usage for linux systems
        corememusage = float(resource.getrusage(resource.RUSAGE_SELF).ru_maxrss) / (1024 * 1024)
        subprocmemusage = float(resource.getrusage(resource.RUSAGE_CHILDREN).ru_maxrss) / (1024 * 1024)
        self._finish = round(corememusage + subprocmemusage, 2)

    @property
    def memoryused(self):
        if self._finish is None or self._start is None:
            return '?'
        return self._finish


def resolve_arg(arg):
    if not isinstance(arg, pypeliner.arguments.Arg):
        return None, False
    return arg.resolve(), True


class JobCallable(object):
    """ Callable function and args to be given to exec queue """

    def __init__(self, id, func, argset, arglist, storage, store_dir, logs_dir, ctx):
        self.storage = storage
        self.id = id
        self.ctx = ctx
        self.func = func
        self.argset = argset
        self.arglist = arglist
        self.started = False
        self.finished = False
        self.displaycommand = '?'
        self.store_dir = store_dir
        self.logs_dir = logs_dir
        self.stdout_filename = os.path.join(logs_dir, 'job.out')
        self.stderr_filename = os.path.join(logs_dir, 'job.err')
        self.stdout_storage = self.storage.create_store(self.stdout_filename)
        self.stderr_storage = self.storage.create_store(self.stderr_filename)
        self.job_timer = JobTimer()
        self.job_mem_tracker = JobMemoryTracker()
        timeout = self.ctx.get("timeout", None) if ctx else None
        self.job_time_out = JobTimeOut(timeout)
        self.hostname = '?'
        self.pypeliner_globals = pypeliner.helpers.GlobalState.get_all()

    @property
    def duration(self):
        return self.job_timer.duration

    @property
    def memoryused(self):
        return self.job_mem_tracker.memoryused

    def log_text(self):
        text = '--- stdout ---\n'
        try:
            with open(self.stdout_filename, 'r') as job_stdout:
                data = deque(job_stdout, 100)
                if len(data) == 100:
                    text += '...\n'
                for line in data:
                    text += line
        except IOError:
            text += 'missing file ' + self.stdout_filename + '\n'
        text += '--- stderr ---\n'
        try:
            with open(self.stderr_filename, 'r') as job_stderr:
                data = deque(job_stderr, 100)
                if len(data) == 100:
                    text += '...\n'
                for line in data:
                    text += line
        except IOError:
            text += 'missing file ' + self.stderr_filename + '\n'
        return text

    def allocate(self):
        for arg in self.arglist:
            arg.allocate()

    def pull(self):
        for arg in self.arglist:
            arg.pull()

    def push(self):
        for arg in self.arglist:
            arg.push()

    def get_displaycommand(self, func, callset):
        if func == pypeliner.commandline.execute:
            displaycommand = '"' + ' '.join(str(arg) for arg in callset.args) + '"'
        else:
            callset_args = ', '.join(repr(arg) for arg in callset.args)
            callset_kwargs = ', '.join(key + '=' + repr(arg) for key, arg in callset.kwargs.items())
            displaycommand = func.__module__ + '.' + func.__name__ + '(' + callset_args + ', ' + callset_kwargs + ')'
        return displaycommand

    def __call__(self):
        pypeliner.helpers.GlobalState.update_all(self.pypeliner_globals)
        self.stdout_storage.allocate()
        self.stderr_storage.allocate()
        with open(self.stdout_storage.write_filename, 'w') as stdout_file, open(self.stderr_storage.write_filename,
                                                                                'w') as stderr_file:
            old_stdout, old_stderr = sys.stdout, sys.stderr
            sys.stdout, sys.stderr = stdout_file, stderr_file
            try:
                pypeliner.helpers.makedirs(self.logs_dir)
                pypeliner.helpers.makedirs(self.store_dir)
                self.started = True
                func = self.func
                if isinstance(func, six.string_types):
                    func = pypeliner.helpers.import_function(func)
                callset = pypeliner.deep.deeptransform(self.argset, resolve_arg)
                self.displaycommand = self.get_displaycommand(func, callset)
                self.hostname = socket.gethostname()
                with self.job_timer, self.job_mem_tracker, self.job_time_out:
                    self.allocate()
                    self.pull()
                    ret_value = func(*callset.args, **callset.kwargs)
                    if callset.ret is not None:
                        callset.ret.finalize(ret_value)
                    self.push()
                self.finished = True
            except Exception:
                sys.stderr.write(traceback.format_exc())
            finally:
                sys.stdout, sys.stderr = old_stdout, old_stderr
        self.stdout_storage.push()
        self.stderr_storage.push()

    def collect_logs(self):
        self.stdout_storage.allocate()
        self.stderr_storage.allocate()
        try:
            self.stdout_storage.pull()
        except pypeliner.storage.InputMissingException:
            pass
        try:
            self.stderr_storage.pull()
        except pypeliner.storage.InputMissingException:
            pass

    def finalize(self, job):
        for arg in self.arglist:
            arg.update(job)
        if job.check_require_regenerate():
            job.workflow.regenerate()
        job.workflow.complete_job(job)


def _setobj_helper(value):
    return value


class SetObjDefinition(JobDefinition):
    def __init__(self, name, axes, ctx, obj, value):
        super(SetObjDefinition, self).__init__(
            name, axes, ctx, _setobj_helper,
            CallSet(ret=obj, args=(value,)))

    def create_job_instances(self, workflow, db):
        for node in db.nodemgr.retrieve_nodes(self.axes):
            yield SetObjInstance(self, workflow, db, node)


class SetObjInstance(JobInstance):
    """ Represents a sub workflow. """

    def __init__(self, job_def, workflow, db, node):
        super(SetObjInstance, self).__init__(job_def, workflow, db, node)
        obj_node = pypeliner.identifiers.create_undefined_node(job_def.argset.ret.axes)
        obj_res = pypeliner.resources.Resource(job_def.argset.ret.name, node + obj_node)
        self.obj_displayname = obj_res.build_displayname(workflow.node)


class SubWorkflowDefinition(JobDefinition):
    def create_job_instances(self, workflow, db):
        for node in db.nodemgr.retrieve_nodes(self.axes):
            yield SubWorkflowInstance(self, workflow, db, node)


class SubWorkflowInstance(JobInstance):
    """ Represents a sub workflow. """
    direct_write = True

    def __init__(self, job_def, workflow, db, node):
        super(SubWorkflowInstance, self).__init__(job_def, workflow, db, node)

    def create_callable(self):
        return WorkflowCallable(
            self.id, self.job_def.func, self.argset, self.arglist,
            self.db.file_storage, self.store_dir, self.logs_dir,
            self.job_def.ctx)

    def init_inputs_outputs(self):
        self.inputs = list()
        self.outputs = list()
        merge_inputs = list()
        split_outputs = list()
        for arg in self.arglist:
            # do not keep track of return object from subworkflow. this will allow pypeliner
            # to skip subworkflow even if we dont have return object on disk as long as all
            # inputs and outputs are up to date. input_resources and output_resources are only
            # used to track whether job is up to date.
            if not isinstance(arg, pypeliner.arguments.Arg):
                continue
            if isinstance(arg, pypeliner.arguments.OutputWorkflowArg):
                continue
            self.inputs.extend(arg.get_inputs())
            self.outputs.extend(arg.get_outputs())
            merge_inputs.extend(arg.get_merge_inputs())
            split_outputs.extend(arg.get_split_outputs())
        # A dependency that is both a merge input and split output
        # is only an output
        split_output_ids = set([a.id for a in split_outputs])
        for input_ in merge_inputs:
            if input_.id not in split_output_ids:
                self.inputs.append(input_)
        self.outputs.extend(split_outputs)
        for node_input in self.db.nodemgr.get_node_inputs(self.node):
            self.inputs.append(node_input)


class WorkflowCallable(JobCallable):
    def allocate(self):
        self.argset.ret.allocate()

    def push(self):
        self.argset.ret.push()

    def pull(self):
        pass

    def __call__(self):
        pypeliner.workflow.parent_ctx = self.ctx.copy()
        # TODO: local shouldnt be inherited by workflows
        # find a better solution to this
        if 'local' in pypeliner.workflow.parent_ctx:
            del pypeliner.workflow.parent_ctx['local']
        super(WorkflowCallable, self).__call__()
        pypeliner.workflow.parent_ctx = None

    def finalize(self, job):
        self.argset.ret.allocate()
        self.argset.ret.pull()
        workflow_def = self.argset.ret.get_obj()
        if not isinstance(workflow_def, pypeliner.workflow.Workflow):
            job.workflow._logger.error(
                'subworkflow ' + job.displayname + ' did not return a workflow\n' + self.log_text(),
                extra={"id": job.displayname, "type": "subworkflow", "status": "error", 'task_name': job.id[1]})
            raise IncompleteWorkflowException()
        if workflow_def.empty:
            job.workflow._logger.error(
                'subworkflow ' + job.displayname + ' returned an empty workflow\n' + self.log_text(),
                extra={"id": job.displayname, "type": "subworkflow", "status": "empty", 'task_name': job.id[1]})
            raise IncompleteWorkflowException()
        job.workflow.add_subworkflow(job, workflow_def)



import json
import logging
import os
import subprocess
from string import Formatter

import pypeliner.delegator
import pypeliner.execqueue.local
import pypeliner.execqueue.qcmd
import pypeliner.execqueue.utils
import pypeliner.helpers
import time


class NativespecFormatter(dict):
    def __missing__(self, key):
        if '=' in key:
            assert len(key.split('=')) == 2
            key, default = key.split('=')
            return self.get(key, default)
        else:
            raise KeyError('Key {} missing in job context'.format(key))


class QsubJob(object):
    """ Encapsulate a running job created using a queueing system's
    qsub submit command called using subprocess
    """

    def __init__(self, ctx, name, sent, temps_dir, modules, qsub_bin, native_spec):
        self.name = name
        self.logger = logging.getLogger('pypeliner.execqueue')
        self.delegated = pypeliner.delegator.Delegator(sent, os.path.join(temps_dir, 'job.dgt'), modules)
        self.command = self.delegated.initialize()
        self.debug_filenames = dict()
        self.debug_filenames['job stdout'] = os.path.join(temps_dir, 'job.out')
        self.debug_filenames['job stderr'] = os.path.join(temps_dir, 'job.err')
        self.debug_filenames['submit stdout'] = os.path.join(temps_dir, 'submit.out')
        self.debug_filenames['submit stderr'] = os.path.join(temps_dir, 'submit.err')
        self.debug_files = []
        self.script_filename = os.path.join(temps_dir, 'submit.sh')
        with open(self.script_filename, 'w') as script_file:
            script_file.write("#!/bin/sh\n")
            script_file.write(' '.join(self.command) + '\n')
        pypeliner.helpers.set_executable(self.script_filename)
        self.submit_command = self.create_submit_command(
            ctx, self.name, self.script_filename, qsub_bin, native_spec,
            self.debug_filenames['job stdout'],
            self.debug_filenames['job stderr']
        )
        try:
            self.debug_files.append(open(self.debug_filenames['submit stdout'], 'w'))
            self.debug_files.append(open(self.debug_filenames['submit stderr'], 'w'))
            self.process = subprocess.Popen(self.submit_command, stdout=self.debug_files[0], stderr=self.debug_files[1])
        except OSError as e:
            self.close_debug_files()
            error_text = self.name + ' submit failed\n'
            error_text += '-' * 10 + ' submit command ' + '-' * 10 + '\n'
            error_text += ' '.join(self.submit_command) + '\n'
            if type(e) == OSError and e.errno == 2:
                error_text += str(e) + ', ' + self.submit_command[0] + ' not found\n'
            else:
                error_text += str(e) + '\n'
            error_text += pypeliner.execqueue.utils.log_text(self.debug_filenames)
            self.logger.error(error_text,
                              extra={"id": self.name, "status": "submit_fail", "type": "job"})
            raise pypeliner.execqueue.base.SubmitError()

    def create_submit_command(
            self, ctx, name, script_filename, qsub_bin, native_spec,
            stdout_filename, stderr_filename
    ):
        qsub = [qsub_bin]
        qsub += ['-sync', 'y', '-b', 'y']
        qsub += Formatter().vformat(native_spec, (), NativespecFormatter(**ctx))
        qsub += ['-N', pypeliner.execqueue.utils.qsub_format_name(name)]
        qsub += ['-o', stdout_filename]
        qsub += ['-e', stderr_filename]
        qsub += [script_filename]
        return qsub

    def close_debug_files(self):
        for file in self.debug_files:
            file.close()

    def finalize(self, returncode):
        self.close_debug_files()
        self.received = self.delegated.finalize()
        if returncode != 0 or self.received is None or not self.received.started:
            error_text = self.name + ' failed to complete\n'
            error_text += '-' * 10 + ' delegator command ' + '-' * 10 + '\n'
            error_text += ' '.join(self.command) + '\n'
            error_text += '-' * 10 + ' submit command ' + '-' * 10 + '\n'
            error_text += ' '.join(self.submit_command) + '\n'
            error_text += 'return code {0}\n'.format(returncode)
            error_text += pypeliner.execqueue.utils.log_text(self.debug_filenames)
            self.logger.error(error_text,
                              extra={"id": self.name, "status": "fail", "type": "job"})
            raise pypeliner.execqueue.base.ReceiveError()


class QsubJobQueue(pypeliner.execqueue.subproc.SubProcessJobQueue):
    """ Queue of qsub jobs """

    def __init__(self, modules=None, native_spec=None):
        super(QsubJobQueue, self).__init__(modules)
        self.qsub_bin = pypeliner.helpers.which('qsub')
        self.native_spec = native_spec
        self.local_queue = pypeliner.execqueue.local.LocalJobQueue(modules)

    def create(self, ctx, name, sent, temps_dir):
        if ctx.get('local', False):
            return pypeliner.execqueue.local.LocalJob(ctx, name, sent, temps_dir, self.modules)
        else:
            return QsubJob(ctx, name, sent, temps_dir, self.modules, self.qsub_bin, self.native_spec)


class AsyncQsubJob(object):
    """ Encapsulate a running job created using a queueing system's
    qsub submit command called using subprocess, and polled using qstat
    """

    def __init__(
            self, ctx, name, sent, temps_dir, modules, qenv, native_spec,
            qstat_job_status, qsub_wrapper, qacct_wrapper
    ):
        self.name = name
        self.qenv = qenv
        self.qstat_job_status = qstat_job_status
        self.qsub_job_id = None
        self.qacct = None
        self.logger = logging.getLogger('pypeliner.execqueue')

        self.delegated = pypeliner.delegator.Delegator(sent, os.path.join(temps_dir, 'job.dgt'), modules)
        self.command = self.delegated.initialize()

        self.debug_filenames = dict()
        self.debug_filenames['job stdout'] = os.path.join(temps_dir, 'job.out')
        self.debug_filenames['job stderr'] = os.path.join(temps_dir, 'job.err')
        self.debug_filenames['submit stdout'] = os.path.join(temps_dir, 'submit.out')
        self.debug_filenames['submit stderr'] = os.path.join(temps_dir, 'submit.err')
        self.debug_filenames['qacct stdout'] = os.path.join(temps_dir, 'qacct.out')
        self.debug_filenames['qacct stderr'] = os.path.join(temps_dir, 'qacct.err')
        for filename in self.debug_filenames.values():
            pypeliner.helpers.saferemove(filename)

        self.script_filename = os.path.join(temps_dir, 'submit.sh')
        with open(self.script_filename, 'w') as script_file:
            script_file.write("#!/bin/sh\n")
            script_file.write(' '.join(self.command) + '\n')
        pypeliner.helpers.set_executable(self.script_filename)

        self.qsub = qsub_wrapper(
            self.qenv, ctx, name, self.script_filename, native_spec, self.debug_filenames['job stdout'],
            self.debug_filenames['job stderr'], self.debug_filenames['submit stdout'],
            self.debug_filenames['submit stderr'])

        self.submit_command = self.qsub.submit_command
        self.qsub_job_id = self.qsub.submit_job()

        self.qsub_time = time.time()

        self.qacct = qacct_wrapper(
            self.qenv, self.qsub_job_id,
            self.debug_filenames['qacct stdout'],
            self.debug_filenames['qacct stderr'],
        )

    @property
    def finished(self):
        """ Get job finished boolean.
        """
        if not self.qstat_job_status.finished(self.qsub_job_id, self.qsub_time) and \
                not self.qstat_job_status.errors(self.qsub_job_id):
            return False

        if self.qstat_job_status.errors(self.qsub_job_id):
            try:
                self.qacct.check()
            except pypeliner.execqueue.qcmd.QacctError:
                pass

        return True

    def finalize(self):
        """ Finalize a job after remote run.
        """
        assert self.finished

        if self.qstat_job_status.errors(self.qsub_job_id):
            self.delete()
            raise pypeliner.execqueue.base.ReceiveError(self.create_error_text('job error'))

        if self.qacct.results is not None and self.qacct.results['exit_status'] != '0':
            raise pypeliner.execqueue.base.ReceiveError(self.create_error_text('qsub error'))

        self.received = self.delegated.finalize()

        if self.received is None or not self.received.started:
            raise pypeliner.execqueue.base.ReceiveError(self.create_error_text('receive error'))

    def delete(self):
        """ Delete job from queue.
        """
        try:
            subprocess.check_call([self.qenv.qdel_bin, self.qsub_job_id])
        except:
            self.logger.error('Unable to delete {}'.format(self.qsub_job_id),
                              extra={"id": self.name, "status": "could_not_delete_job", "type": "job"})

    def create_error_text(self, desc):
        """ Create error text string.

        Args:
            desc (str): error description

        Returns:
            str: multi-line error text
        """

        error_text = list()

        error_text += [desc + ' for job: ' + self.name]

        if self.qsub_job_id is not None:
            error_text += ['qsub id: ' + str(self.qsub_job_id)]

        error_text += ['delegator command: ' + ' '.join(self.command)]
        error_text += ['submit command: ' + ' '.join(self.submit_command)]

        if self.qacct is not None and self.qacct.results is not None:
            error_text += ['memory consumed: ' + self.qacct.results['maxvmem']]
            error_text += ['job exit status: ' + self.qacct.results['exit_status']]

        for debug_type, debug_filename in self.debug_filenames.items():
            if not os.path.exists(debug_filename):
                error_text += [debug_type + ': missing']
                continue
            with open(debug_filename, 'r') as debug_file:
                error_text += [debug_type + ':']
                for line in debug_file:
                    error_text += ['\t' + line.rstrip()]

        return '\n'.join(error_text)


class AsyncQsubJobQueue(pypeliner.execqueue.base.JobQueue):
    """ Class for a queue of jobs run using subprocesses.  Maintains
    a list of running jobs, with the ability to wait for jobs and return
    completed jobs.  Requires override of the create method.
    """

    def __init__(self, modules=None, **kwargs):
        self.modules = modules
        self.native_spec = kwargs['native_spec']
        self.qstat = pypeliner.execqueue.qcmd.QstatJobStatus(self.qenv)
        self.jobs = dict()
        self.name_islocal = dict()
        self.local_queue = pypeliner.execqueue.local.LocalJobQueue(modules)

    def __enter__(self):
        self.local_queue.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.local_queue.__exit__(exc_type, exc_value, traceback)
        for job in self.jobs.values():
            job.delete()

    def create(self, ctx, name, sent, temps_dir):
        return AsyncQsubJob(ctx, name, sent, temps_dir, self.modules, self.qenv, self.native_spec, self.qstat,
                            pypeliner.execqueue.qcmd.QsubWrapper, pypeliner.execqueue.qcmd.QacctWrapper)

    def send(self, ctx, name, sent, temps_dir):
        if ctx.get('local', False):
            self.local_queue.send(ctx, name, sent, temps_dir)
        else:
            self.jobs[name] = self.create(ctx, name, sent, temps_dir)

    def wait(self):
        while True:
            if not self.local_queue.empty:
                name = self.local_queue.wait(immediate=True)
                if name is not None:
                    self.name_islocal[name] = True
                    return name
            for name, job in self.jobs.items():
                if job.finished:
                    return name
            self.qstat.update(self.jobs)

    def receive(self, name):
        if self.name_islocal.pop(name, False):
            return self.local_queue.receive(name)
        job = self.jobs.pop(name)
        job.finalize()
        return job.received

    @property
    def length(self):
        return len(self.jobs) + self.local_queue.length

    @property
    def empty(self):
        return self.length == 0

    @property
    def qenv(self):
        return pypeliner.execqueue.qcmd.QEnv()


class PbsQstatJobStatus(pypeliner.execqueue.qcmd.QstatJobStatus):
    """ Statuses of jobs on a pbs cluster """

    def finished(self):
        return 'c' in self.cached_job_status.get(self.job_id, 'c')

    def errors(self, job_id):
        return False


class PbsJobQueue(AsyncQsubJobQueue):
    """ Queue of jobs running on a pbs cluster """

    def __init__(self, modules=None, **kwargs):
        super(PbsJobQueue, self).__init__(modules, **kwargs)
        self.qstat = PbsQstatJobStatus(self.qenv)


class LsfQstatJobStatus(pypeliner.execqueue.qcmd.QstatJobStatus):
    """ Statuses of jobs on a lfs cluster """

    def get_qstat_job_status(self, jobs):
        job_status = dict()

        qsub_ids = [val.qsub_job_id for val in jobs.values()]

        qsub_id_groups = [qsub_ids[x:x + 100] for x in range(0, len(qsub_ids), 100)]

        for ids_group in qsub_id_groups:
            cmd = [self.qenv.qstat_bin, '-o', 'JOBID:10 STAT:6', '-json']
            cmd.extend(ids_group)

            qstat_output = subprocess.check_output(cmd).decode()
            qstat_output = json.loads(qstat_output)

            for record in qstat_output['RECORDS']:
                if 'ERROR' in record:
                    continue
                if record['STAT'] not in ['PEND', 'WAIT', 'PROV', 'RUN']:
                    continue
                try:
                    job_status[record['JOBID']] = record['STAT'].lower()
                except IndexError:
                    continue
        return job_status

    def errors(self, qsub_job_id):
        if self.cached_job_status is None:
            return None
        # possible error states are POST_ERR state if it fails, or {P|U|S}SUSP states
        return 'ERR' in self.cached_job_status.get(qsub_job_id, '') or \
            'SUSP' in self.cached_job_status.get(qsub_job_id, '')


class SlurmQstatJobStatus(pypeliner.execqueue.qcmd.QstatJobStatus):
    """ Statuses of jobs on a lfs cluster """

    def get_qstat_job_status(self, jobs):
        job_status = dict()

        qsub_ids = [val.qsub_job_id for val in jobs.values()]

        qsub_id_groups = [qsub_ids[x:x + 100] for x in range(0, len(qsub_ids), 100)]

        for ids_group in qsub_id_groups:
            cmd = [self.qenv.qstat_bin, '-j', ','.join([str(v) for v in ids_group])]
            qstat_output = subprocess.check_output(cmd).decode().split('\n')
            assert len(qstat_output) > 1
            header = {v: i for i, v in enumerate(qstat_output[0].strip().split())}

            for line in qstat_output[1:]:
                line = line.strip().split()
                if len(line) == 0:
                    continue
                status = line[header['ST']]
                jobid = int(line[header['JOBID']])

                if status not in ['CF', 'CG', 'PD', 'R', 'RD', 'RF', 'RH', 'RQ', 'RS', 'RV', 'SI', 'SE', 'SO']:
                    continue

                job_status[jobid] = status

        return job_status

    def errors(self, qsub_job_id):
        if self.cached_job_status is None:
            return None
        # possible error states are POST_ERR state if it fails, or {P|U|S}SUSP states
        return 'F' in self.cached_job_status.get(qsub_job_id, '') or \
            'S' in self.cached_job_status.get(qsub_job_id, '') or \
            'OOM' in self.cached_job_status.get(qsub_job_id, '') or \
            'BF' in self.cached_job_status.get(qsub_job_id, '') or \
            'CA' in self.cached_job_status.get(qsub_job_id, '') or \
            'DL' in self.cached_job_status.get(qsub_job_id, '') or \
            'SE' in self.cached_job_status.get(qsub_job_id, '') or \
            'ST' in self.cached_job_status.get(qsub_job_id, '') or \
            'TO' in self.cached_job_status.get(qsub_job_id, '')


class LsfJobQueue(AsyncQsubJobQueue):
    """ Queue of jobs running on a lfs cluster """

    def __init__(self, modules=None, **kwargs):
        super(LsfJobQueue, self).__init__(modules, **kwargs)
        self.qstat = LsfQstatJobStatus(self.qenv)

    def create(self, ctx, name, sent, temps_dir):
        """create job"""
        return AsyncQsubJob(ctx, name, sent, temps_dir, self.modules, self.qenv, self.native_spec, self.qstat,
                            pypeliner.execqueue.qcmd.LsfsubWrapper, pypeliner.execqueue.qcmd.LsfacctWrapper)

    @property
    def qenv(self):
        return pypeliner.execqueue.qcmd.LsfEnv()


class SlurmJobQueue(AsyncQsubJobQueue):
    """ Queue of jobs running on a lfs cluster """

    def __init__(self, modules=None, **kwargs):
        super(SlurmJobQueue, self).__init__(modules, **kwargs)
        self.qstat = SlurmQstatJobStatus(self.qenv)

    def create(self, ctx, name, sent, temps_dir):
        """create job"""
        return AsyncQsubJob(ctx, name, sent, temps_dir, self.modules, self.qenv, self.native_spec, self.qstat,
                            pypeliner.execqueue.qcmd.SlurmsubWrapper, pypeliner.execqueue.qcmd.SlurmacctWrapper)

    @property
    def qenv(self):
        return pypeliner.execqueue.qcmd.SlurmEnv()



import cmd
import fnmatch
import logging


class BasicRunSkip(object):
    def __init__(self, repopulate=False, rerun=False):
        self.repopulate = repopulate
        self.rerun = rerun

    def __call__(self, job):
        if job.runskip_request == "skip":
            return False, 'skip requested'
        if self.rerun:
            return True, 'rerun requested\n' + job.explain_out_of_date()
        if job.out_of_date():
            return True, job.explain_out_of_date()
        if job.is_required_downstream:
            return True, 'required downstream\n' + job.explain_out_of_date()
        if self.repopulate and job.output_missing():
            return True, 'repopulate requested, missing output\n' + job.explain_out_of_date()
        return False, job.explain_out_of_date()

    def close(self):
        pass


class SentinalRunSkip(object):
    def __init__(self, rerun=False):
        self.rerun = rerun

    def __call__(self, job):
        if job.runskip_request == "skip":
            return False, 'skip requested'
        if self.rerun:
            return True, 'rerun requested\n' + job.explain_out_of_date()
        if job.already_run():
            return False, 'already run'
        return True, 'not yet run'

    def close(self):
        pass


class PatternMatcher(object):
    def __init__(self):
        self._patterns = list()

    def get(self, value):
        for key, match in self._patterns:
            if fnmatch.fnmatch(value, match):
                return key

    def add(self, key, match):
        self._patterns.append((key, match))

    def get_patterns_str(self):
        patterns = []
        for key, match in self._patterns:
            patterns.append('{} {}'.format(key, match))
        return '\n'.join(patterns)


class RunSkipCmd(cmd.Cmd):
    def __init__(self, patterns):
        cmd.Cmd.__init__(self)
        self.patterns = patterns
        self.command = ''
        self.success = False
        self.prompt = 'pypeliner$ '

    def do_command(self, command, line):
        if len(line) > 0:
            self.patterns.add(command, line)
        else:
            self.command = command
        self.success = True
        return True

    def do_run(self, line):
        return self.do_command('run', line)

    def help_run(self):
        print ('run the job')

    def do_skip(self, line):
        return self.do_command('skip', line)

    def help_skip(self):
        print ('skip the job')

    def do_verify(self, line):
        return self.do_command('verify', line)

    def help_verify(self):
        print ('skip the job it doesnt need to be run')

    def do_touch(self, line):
        return self.do_command('touch', line)

    def help_touch(self):
        print ('touch outputs and skip the job')

    def do_default(self, line):
        return self.do_command('default', line)

    def help_default(self):
        print ('default behaviour')


class InteractiveRunSkip(object):
    def __init__(self, default):
        self.patterns = PatternMatcher()
        self.default = default
        self._logger = logging.getLogger('pypeliner.scheduler.runskip')

    def __call__(self, job):
        default_is_run_required, default_explaination = self.default(job)

        self._logger.info(
            'job ' + job.displayname + ' default run: ' + str(
                default_is_run_required) + ' default explanation: ' + default_explaination,
            extra={
                'job_name': job.displayname, "explanation": default_explaination,
                'task_name': job.id[1]
            }
        )

        runskip_cmd = RunSkipCmd(self.patterns)

        while True:
            if runskip_cmd.command == 'run' or self.patterns.get(job.displayname) == 'run':
                return True, 'run requested'

            if runskip_cmd.command == 'skip' or self.patterns.get(job.displayname) == 'skip':
                return False, 'skip requested'

            if runskip_cmd.command == 'verify' or self.patterns.get(job.displayname) == 'verify':
                if not default_is_run_required:
                    return False, 'verified and skipped'

            if runskip_cmd.command == 'touch' or self.patterns.get(job.displayname) == 'touch':
                job.touch_outputs()
                return False, 'touch requested'

            if runskip_cmd.command == 'default' or self.patterns.get(job.displayname) == 'default':
                return default_is_run_required, 'default requested'

            runskip_cmd.cmdloop()
            if not runskip_cmd.success:
                raise Exception('failed to obtain user input')

    def close(self):
        self._logger.info('run skip commands:\n' + self.patterns.get_patterns_str())




from __future__ import absolute_import
import drmaa
import logging
import os
import time

import pypeliner.delegator
import pypeliner.helpers
import pypeliner.execqueue.base
import pypeliner.execqueue.local
import pypeliner.execqueue.utils


decode_status = {
    drmaa.JobState.UNDETERMINED: 'process status cannot be determined',
    drmaa.JobState.QUEUED_ACTIVE: 'job is queued and active',
    drmaa.JobState.SYSTEM_ON_HOLD: 'job is queued and in system hold',
    drmaa.JobState.USER_ON_HOLD: 'job is queued and in user hold',
    drmaa.JobState.USER_SYSTEM_ON_HOLD: 'job is queued and in user and system hold',
    drmaa.JobState.RUNNING: 'job is running',
    drmaa.JobState.SYSTEM_SUSPENDED: 'job is system suspended',
    drmaa.JobState.USER_SUSPENDED: 'job is user suspended',
    drmaa.JobState.DONE: 'job finished normally',
    drmaa.JobState.FAILED: 'job finished, but failed'
}


class DrmaaJob(object):
    """ Encapsulate a running job created using drmaa
    """

    def __init__(self, ctx, name, sent, temps_dir, modules, native_spec, session):
        self.name = name
        self.native_spec = native_spec
        self.session = session
        self.temps_dir = temps_dir
        self.logger = logging.getLogger('pypeliner.execqueue')
        self.delegated = pypeliner.delegator.Delegator(sent, os.path.join(temps_dir, 'job.dgt'), modules)
        self.command = self.delegated.initialize()

        self.debug_filenames = dict()
        self.debug_filenames['job stdout'] = os.path.join(self.temps_dir, 'job.out')
        self.debug_filenames['job stderr'] = os.path.join(self.temps_dir, 'job.err')
        self.debug_filenames['resources'] = os.path.join(self.temps_dir, 'resources.txt')
        for filename in self.debug_filenames.values():
            pypeliner.helpers.saferemove(filename)

        job_template = self.session.createJobTemplate()
        job_template.remoteCommand = self.command[0]
        job_template.args = self.command[1:]

        native_spec, export_local_env = self._parse_native_spec(native_spec)
        if export_local_env:
            job_template.jobEnvironment = os.environ

        job_template.nativeSpecification = self._create_native_spec(native_spec, ctx)
        job_template.outputPath = ':' + self.debug_filenames['job stdout']
        job_template.errorPath = ':' + self.debug_filenames['job stderr']
        job_template.jobName = pypeliner.execqueue.utils.qsub_format_name(self.name)

        self.job_id = self.session.runJob(job_template)
        self.session.deleteJobTemplate(job_template)

        self.job_info = None

        self.unrecoverable_error = False

    @property
    def finished(self):
        """ Get job finished boolean.
        """
        if self.unrecoverable_error:
            return True

        if self.job_info is None:
            try:
                job_status = self.session.jobStatus(self.job_id)
            except Exception as e:
                self.unrecoverable_error = True
                self.unrecoverable_exception = str(e)
                return True

            if job_status in [drmaa.JobState.QUEUED_ACTIVE, drmaa.JobState.RUNNING, drmaa.JobState.UNDETERMINED]:
                return False

            self.job_info = self.session.wait(self.job_id)

        return True

    def finalize(self):
        assert self.finished

        if self.unrecoverable_error:
            raise pypeliner.execqueue.base.ReceiveError(
                self._create_error_text('unrecoverable error: ' + self.unrecoverable_exception))

        self._write_resource_usage()

        if int(self.job_info.exitStatus) != 0 or int(float(self.job_info.resourceUsage.get('exit_status', 0))) != 0:
            raise pypeliner.execqueue.base.ReceiveError(self._create_error_text('drmma error'))

        self.received = self.delegated.finalize()

        if self.received is None:
            raise pypeliner.execqueue.base.ReceiveError(self._create_error_text('receive error'))

    def _create_error_text(self, desc):
        """ Create error text string.

        Args:
            desc (str): error description

        Returns:
            str: multi-line error text
        """

        error_text = list()

        error_text.append('{0} for job: {1}'.format(desc, self.name))

        error_text.append('qsub id: {0}'.format(self.job_id))

        cmd_str = ' '.join(self.command)

        error_text.append('delegator command: {0}'.format(cmd_str))

        error_text.append('native specification: {}'.format(self.native_spec))

        if self.job_info is not None:
            if self.job_info.wasAborted:
                error_text.append('job was aborted')

            elif self.job_info.hasSignal:
                error_text.append('job finished due to signal {0}.'.format(self.job_info.terminatedSignal))

                if self.job_info.hasCoreDump:
                    error_text.append('a core dump is available for job')

            else:
                error_text.append('job finished with unclear conditions')

            error_text.append('memory consumed: {0}'.format(
                self.job_info.resourceUsage.get('maxvmem',
                                                'unknown, values are: ' + repr(self.job_info.resourceUsage))))

            error_text.append('job exit status: {0}'.format(self.job_info.exitStatus))

        else:
            error_text.append('no job info available')

        for debug_type, debug_filename in self.debug_filenames.items():
            if not os.path.exists(debug_filename):
                error_text += [debug_type + ': missing']

                continue

            with open(debug_filename, 'r') as debug_file:
                error_text += [debug_type + ':']

                for line in debug_file:
                    error_text += ['\t' + line.rstrip()]

        return '\n'.join(error_text)

    def _create_native_spec(self, native_spec, ctx):
        return native_spec.format(**ctx)

    def _parse_native_spec(self, native_spec):
        if '--export-local-env' in native_spec:
            native_spec = native_spec.replace('--export-local-env', '')

            export_local_env = True

        else:
            export_local_env = False

        return native_spec, export_local_env

    def _write_resource_usage(self):
        resources_text = []

        resources_text.append('=' * 80)

        max_len = max([len(x) for x in self.job_info.resourceUsage.keys()])

        for key, value in sorted(self.job_info.resourceUsage.items()):
            num_spaces = max_len - len(key) + 1

            if key in ['acct_maxvmem', 'mem', 'maxvmem', 'vmem']:
                value = '{0}G'.format(round(float(value) / 1e9, 2))

            if key in ['exit_status', 'failed', 'signal', 'slots']:
                value = str(int(float(value)))

            if key in ['end_time', 'start_time', 'submission_time']:
                value = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(float(value)))

            out_str = key + ' ' * num_spaces + value

            resources_text.append(out_str)

        with open(self.debug_filenames['resources'], 'w') as fh:
            fh.write('\n'.join(resources_text))


class DrmaaJobQueue(object):
    """ Maintain a list of running jobs executed synchronously using
    drmaa, with the ability to wait for jobs and return completed jobs
    """

    def __init__(self, modules=None, native_spec=None, **kwargs):
        self.modules = modules

        self.native_spec = native_spec

        self.jobs = dict()

        self.name_islocal = dict()

        self.local_queue = pypeliner.execqueue.local.LocalJobQueue(modules=modules)

    def __enter__(self):
        self.local_queue.__enter__()

        self.session = drmaa.Session()

        self.session.initialize()

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.local_queue.__exit__(exc_type, exc_value, traceback)

        self.session.control(drmaa.Session.JOB_IDS_SESSION_ALL, drmaa.JobControlAction.TERMINATE)

        self.session.exit()

    def create(self, ctx, name, sent, temps_dir):
        return DrmaaJob(ctx, name, sent, temps_dir, self.modules, self.native_spec, self.session)

    def send(self, ctx, name, sent, temps_dir):
        if ctx.get('local', False):
            self.local_queue.send(ctx, name, sent, temps_dir)

        else:
            self.jobs[name] = self.create(ctx, name, sent, temps_dir)

    def wait(self):
        while True:
            if not self.local_queue.empty:
                name = self.local_queue.wait(immediate=True)

                if name is not None:
                    self.name_islocal[name] = True

                    return name

            while True:
                for name, job in self.jobs.items():
                    if job.finished:
                        return name

            time.sleep(1)

    def receive(self, name):
        if self.name_islocal.pop(name, False):
            return self.local_queue.receive(name)

        job = self.jobs.pop(name)

        job.finalize()

        return job.received

    @property
    def length(self):
        return len(self.jobs) + self.local_queue.length

    @property
    def empty(self):
        return self.length == 0



""" Reworked version of standard libary copy module for transforming
arbitrary python objects.
"""

import sys
import types
import weakref

import six

if sys.version_info[0] < 3:
    from copy_reg import dispatch_table
else:
    from copyreg import dispatch_table

    long = int
    xrange = range


class Error(Exception):
    pass


error = Error  # backward compatibility

__all__ = ["Error", "copy", "deeptransform"]


def deeptransform(x, f, memo=None, _nil=[]):
    """Deep transform operation on arbitrary Python objects.

    See the module's __doc__ string for more info.
    """

    if memo is None:
        memo = {}

    d = id(x)
    y = memo.get(d, _nil)
    if y is not _nil:
        return y

    cls = type(x)

    y, transformed = f(x)
    if not transformed:
        copier = _deeptransform_dispatch.get(cls)
        if copier:
            y = copier(x, f, memo)
        else:
            try:
                issc = issubclass(cls, type)
            except TypeError:  # cls is not a class (old Boost; see SF #502085)
                issc = 0
            if issc:
                y = _deeptransform_atomic(x, f, memo)
            else:
                copier = getattr(x, "__deeptransform__", None)
                if copier:
                    y = copier(memo)
                else:
                    reductor = dispatch_table.get(cls)
                    if reductor:
                        rv = reductor(x)
                    else:
                        reductor = getattr(x, "__reduce_ex__", None)
                        if reductor:
                            rv = reductor(2)
                        else:
                            reductor = getattr(x, "__reduce__", None)
                            if reductor:
                                rv = reductor()
                            else:
                                raise Error(
                                    "un(deep)copyable object of type %s" % cls)
                    y = _reconstruct(x, f, rv, 1, memo)

    memo[d] = y
    _keep_alive(x, f, memo)  # Make sure x lives at least as long as d
    return y


_deeptransform_dispatch = d = {}


def _deeptransform_atomic(x, f, memo):
    return x


d[type(None)] = _deeptransform_atomic
d[type(Ellipsis)] = _deeptransform_atomic
d[int] = _deeptransform_atomic
d[float] = _deeptransform_atomic
d[bool] = _deeptransform_atomic
try:
    d[complex] = _deeptransform_atomic
except NameError:
    pass
d[str] = _deeptransform_atomic
try:
    d[unicode] = _deeptransform_atomic
except NameError:
    pass
try:
    d[types.CodeType] = _deeptransform_atomic
except AttributeError:
    pass
d[type] = _deeptransform_atomic
d[xrange] = _deeptransform_atomic
# d[types.ClassType] = _deeptransform_atomic
d[types.BuiltinFunctionType] = _deeptransform_atomic
d[types.FunctionType] = _deeptransform_atomic
d[weakref.ref] = _deeptransform_atomic


def _deeptransform_list(x, f, memo):
    y = []
    memo[id(x)] = y
    for a in x:
        y.append(deeptransform(a, f, memo))
    return y


d[list] = _deeptransform_list


def _deeptransform_tuple(x, f, memo):
    y = []
    for a in x:
        y.append(deeptransform(a, f, memo))
    d = id(x)
    try:
        return memo[d]
    except KeyError:
        pass
    for i in range(len(x)):
        if x[i] is not y[i]:
            y = tuple(y)
            break
    else:
        y = x
    memo[d] = y
    return y


d[tuple] = _deeptransform_tuple


def _deeptransform_dict(x, f, memo):
    y = {}
    memo[id(x)] = y
    for key, value in x.items():
        y[deeptransform(key, f, memo)] = deeptransform(value, f, memo)
    return y


d[dict] = _deeptransform_dict


def _deeptransform_method(x, f, memo):  # Copy instance methods
    try:
        return type(x)(x.__func__, deeptransform(x.__self__, f, memo), x.__class__)
    except Exception as e:
        return type(x)(x.__func__, deeptransform(x.__self__, f, memo))


_deeptransform_dispatch[types.MethodType] = _deeptransform_method


def _keep_alive(x, f, memo):
    """Keeps a reference to the object x in the memo.

    Because we remember objects by their id, we have
    to assure that possibly temporary objects are kept
    alive by referencing them.
    We store a reference at the id of the memo, which should
    normally not be used unless someone tries to deeptransform
    the memo itself...
    """
    try:
        memo[id(memo)].append(x)
    except KeyError:
        # aha, this is the first one :-)
        memo[id(memo)] = [x]


def _deeptransform_inst(x, f, memo):
    if hasattr(x, '__deeptransform__'):
        return x.__deeptransform__(memo)
    if hasattr(x, '__getinitargs__'):
        args = x.__getinitargs__()
        args = deeptransform(args, f, memo)
        y = x.__class__(*args)
    else:
        y = _EmptyClass()
        y.__class__ = x.__class__
    memo[id(x)] = y
    if hasattr(x, '__getstate__'):
        state = x.__getstate__()
    else:
        state = x.__dict__
    state = deeptransform(state, f, memo)
    if hasattr(y, '__setstate__'):
        y.__setstate__(state)
    else:
        y.__dict__.update(state)
    return y


if sys.version_info[0] < 3:
    d[types.InstanceType] = _deeptransform_inst


def _reconstruct(x, f, info, deep, memo=None):
    if isinstance(info, six.string_types):
        return x
    assert isinstance(info, tuple)
    if memo is None:
        memo = {}
    n = len(info)
    assert n in (2, 3, 4, 5)
    callable, args = info[:2]
    if n > 2:
        state = info[2]
    else:
        state = None
    if n > 3:
        listiter = info[3]
    else:
        listiter = None
    if n > 4:
        dictiter = info[4]
    else:
        dictiter = None
    if deep:
        args = deeptransform(args, f, memo)
    y = callable(*args)
    memo[id(x)] = y

    if state is not None:
        if deep:
            state = deeptransform(state, f, memo)
        if hasattr(y, '__setstate__'):
            y.__setstate__(state)
        else:
            if isinstance(state, tuple) and len(state) == 2:
                state, slotstate = state
            else:
                slotstate = None
            if state is not None:
                y.__dict__.update(state)
            if slotstate is not None:
                for key, value in slotstate.items():
                    setattr(y, key, value)

    if listiter is not None:
        for item in listiter:
            if deep:
                item = deeptransform(item, f, memo)
            y.append(item)
    if dictiter is not None:
        for key, value in dictiter:
            if deep:
                key = deeptransform(key, f, memo)
                value = deeptransform(value, f, memo)
            y[key] = value
    return y


del d

del types


# Helper for instance creation without calling __init__
class _EmptyClass:
    pass



import logging
import os
import subprocess

import pypeliner.delegator
import pypeliner.execqueue.base
import pypeliner.execqueue.subproc
import pypeliner.execqueue.utils
import pypeliner

class LocalJob(object):
    """ Encapsulate a running job called locally by subprocess """
    def __init__(self, ctx, name, sent, temps_dir, modules):
        self.name = name
        self.logger = logging.getLogger('pypeliner.execqueue')
        self.delegated = pypeliner.delegator.Delegator(sent, os.path.join(temps_dir, 'job.dgt'), modules)
        self.command = self.delegated.initialize()
        self.debug_filenames = dict()
        self.debug_filenames['job stdout'] = os.path.join(temps_dir, 'job.out')
        self.debug_filenames['job stderr'] = os.path.join(temps_dir, 'job.err')
        self.debug_files = []
        try:
            self.debug_files.append(open(self.debug_filenames['job stdout'], 'w'))
            self.debug_files.append(open(self.debug_filenames['job stderr'], 'w'))
            self.process = subprocess.Popen(self.command, stdout=self.debug_files[0], stderr=self.debug_files[1])
        except OSError as e:
            self.close_debug_files()
            error_text = self.name + ' submit failed\n'
            error_text += '-' * 10 + ' delegator command ' + '-' * 10 + '\n'
            error_text += ' '.join(self.command) + '\n'
            error_text += str(e) + '\n'
            error_text += pypeliner.execqueue.utils.log_text(self.debug_filenames)
            self.logger.error(error_text)
            raise pypeliner.execqueue.base.SubmitError()

    def close_debug_files(self):
        for file in self.debug_files:
            file.close()
            
    def finalize(self, returncode):
        self.close_debug_files()
        self.received = self.delegated.finalize()
        if returncode != 0 or self.received is None or not self.received.started:
            error_text = self.name + ' failed to complete\n'
            error_text += '-' * 10 + ' delegator command ' + '-' * 10 + '\n'
            error_text += ' '.join(self.command) + '\n'
            error_text += 'return code {0}\n'.format(returncode)
            error_text += pypeliner.execqueue.utils.log_text(self.debug_filenames)
            self.logger.error(error_text)
            raise pypeliner.execqueue.base.ReceiveError()


class LocalJobQueue(pypeliner.execqueue.subproc.SubProcessJobQueue):
    """ Queue of local jobs """
    def create(self, ctx, name, sent, temps_dir):
        return LocalJob(ctx, name, sent, temps_dir, self.modules)


class LocalRemoteQueue(pypeliner.execqueue.base.JobQueue):
    """ Class for a combined remote and local queue.
    """
    def __init__(self, remote_queue, modules=None):
        self.name_islocal = dict()
        self.local_queue = pypeliner.execqueue.local.LocalJobQueue(modules)
        self.remote_queue = remote_queue

    def __enter__(self):
        self.local_queue.__enter__()
        self.remote_queue.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.local_queue.__exit__(exc_type, exc_value, traceback)
        self.remote_queue.__exit__(exc_type, exc_value, traceback)

    def send(self, ctx, name, sent, temps_dir):
        if not ctx.get('local', False):
            self.remote_queue.send(ctx, name, sent, temps_dir)
        else:
            self.local_queue.send(ctx, name, sent, temps_dir)

    def wait(self):
        while True:
            if not self.local_queue.empty:
                name = self.local_queue.wait(immediate=True)
                if name is not None:
                    self.name_islocal[name] = True
                    return name
            return self.remote_queue.wait()

    def receive(self, name):
        if not self.name_islocal.pop(name, False):
            return self.remote_queue.receive(name)
        else:
            return self.local_queue.receive(name)

    @property
    def length(self):
        return self.local_queue.length + self.remote_queue.length

    @property
    def empty(self):
        return self.length == 0

